/*
 *  Copyright (C) 2000, 2001 Ximian, Inc.
 *
 *  Authors: Joe Shaw <joe@ximian.com>
 *           Michael Zucchi <notzed@ximian.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */

#include <config.h>

#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <sys/types.h>
#include <string.h>
#include <time.h>


#include <glib.h>
#include <gnet/gnet.h>

#include <unistd.h>
#include <fcntl.h>
#include <netinet/in.h>
#include <sys/stat.h>
#include <sys/socket.h>

#include <libgnomeui/gnome-dialog.h>

#include <zlib.h>

#include "gnome-transfer.h"

#define d(x)	 		/* general debug */
#define h(x)    		/* http debug */
#define f(x)			/* ftp debug */
#undef  o                       /* object destruction debug */

#define BLOCK_SIZE 4096

/* eek, this a private data field, but we're going to peek it anyway,
   otherwise async doesn't work */
struct _GTcpSocket
{
  gint sockfd;
  struct sockaddr sa;		/* Why not an InetAddr? */
  guint ref_count;
  GIOChannel* iochannel;
};

static void gnome_transfer_class_init (GnomeTransferClass *klass);
static void gnome_transfer_init       (GnomeTransfer *obj);

/* internal abort function */
static void close_transfer(GnomeTransfer *t);

static GtkObjectClass *gnome_transfer_parent;

#define _PRIVATE(t) (((GnomeTransfer *)t)->priv)

typedef enum   _FTPState             FTPState;
typedef enum   _HTTPState            HTTPState;
typedef enum   _TransferType         TransferType;
typedef struct _GnomeTransferHTTP    GnomeTransferHTTP;
typedef struct _GnomeTransferFTP     GnomeTransferFTP;
typedef struct _GnomeTransferFile    GnomeTransferFile;

enum _FTPState {
	FTP_CONNECT,		/* waiting ftp OK response */
	FTP_USER,		/* sent USER blah */
	FTP_PASS,		/* sent PASS blah */
	FTP_SIZE,		/* sent SIZE blah */
	FTP_TYPE,		/* sent TYPE I */
	FTP_PASV,		/* sent PASV */
	FTP_REST,		/* send RESTart */
	FTP_RETR,		/* sent RETRieve */
	FTP_QUIT,		/* sent QUIT */
};

enum _HTTPState {
	HTTP_STATUS,		/* getting status response */
	HTTP_HEADER,		/* getting header response */
	HTTP_DATA,		/* getting data */
};

enum _TransferType {
	TRANSFER_NONE,
	TRANSFER_HTTP,
	TRANSFER_FTP
};

struct _GnomeTransferHTTP {
	/* HTTP State (HTTP_STATUS, HTTP_HEADER, HTTP_DATA) */
	HTTPState state; /* was httpstate */

	/* HTTP Status Code (200, 404, etc.) */
	int statuscode; /* was httpstatus */

	/* Redirection URL */
	Gurl *redirect;
  
	/* Proxy URL */
	Gurl *proxy;

	/* Proxy authentication challenge */
	char *authrealm;

	/* Proxy authentication password */
	char *authpass;

	/* Client User Agent name */
	char *user_agent;
};

struct _GnomeTransferFTP {
	/* FTP State (FTP_CONNECT, FTP_USER, etc.) */
	FTPState state; /* was ftpstate */

	/* FTP data socket ID */
	GTcpSocketNewAsyncID dataid; /* was datanewid */

	/* FTP data socket */
	GTcpSocket *datasocket;

	/* FTP Anonymous password (usually email) */
	char *anon_password;
};

struct _GnomeTransferFile {
        /* GIOChannel for the file */
        GIOChannel *iochannel;
};

struct _GnomeTransferPrivate {
	/* Type of transfer */
	TransferType type;

	/* Connection function */
	GTcpSocketNewAsyncFunc connect_func; /* was connect */

	/* Data ID */
	GTcpSocketNewAsyncID dataid; /* was newid */

	/* Main Socket */
	GTcpSocket *socket;
	
	/* File Size */
	int size;

	/* Total bytes transfered so far */
	int totalbytes;

	/* Total uncompressed bytes read so far */
	int total_uncompressed;

	/* Start of the transfer offset. 0 means start from the beginning */
	int offset;

	/* We only want to emit the file_info signal once. So, if we pause a
	   transfer and then resume it, we don't want to emit this signal.
	   What we'll do is, when we emit the file_info signal, set this flag
	   on. When a transfer finishes, we'll turn it off. Easy as pie, as
	   they say. */
	gboolean info_emitted;
	
	GnomeTransferHTTP *http;
	GnomeTransferFTP  *ftp;
        GnomeTransferFile *file;
};

enum SIGNALS {
	FILE_INFO,
	FILE_DATA,
	FILE_DONE,
	FILE_PAUSED,
	AUTH_REQUIRED,
	LAST_SIGNAL
};

static guint signals[LAST_SIGNAL] = { 0 };

static gint
insensitive_equal(gconstpointer v1, gconstpointer v2)
{
	const char *s1 = v1;
	const char *s2 = v2;

	if (g_strcasecmp(s1, s2) == 0)
		return TRUE;
	else
		return FALSE;
} /* insensitive_equal */

guint
gnome_transfer_get_type (void)
{
	static guint type = 0;
	
	if (!type) {
		GtkTypeInfo type_info = {
			"GnomeTransfer",
			sizeof (GnomeTransfer),
			sizeof (GnomeTransferClass),
			(GtkClassInitFunc) gnome_transfer_class_init,
			(GtkObjectInitFunc) gnome_transfer_init,
			(GtkArgSetFunc) NULL,
			(GtkArgGetFunc) NULL
		};
		
		type = gtk_type_unique (gtk_object_get_type (), &type_info);
	}
	
	return type;
}

static void
hash_free(gpointer key, gpointer value, gpointer user_data)
{
	g_free(key);
	g_free(value);
} /* hash_free */

static void
gnome_transfer_destroy(GtkObject *obj)
{
	GnomeTransfer *gt = GNOME_TRANSFER(obj);
	GnomeTransferPrivate *p = _PRIVATE(gt);

#if 0
	printf("Object %p being destroyed\n", obj);
#endif

	/* Free stuff in the public structure */
	if (gt->url) {
		g_url_free(gt->url);
		gt->url = NULL;
	}
	if (gt->reason)
		g_free(gt->reason);
	g_hash_table_foreach(gt->headers, hash_free, NULL);
	g_hash_table_destroy(gt->headers);

	/* Free the private structure */
	if (p->socket) {
		gnet_tcp_socket_delete(p->socket);
		p->socket = NULL;
	}
	if (p->http) {
		if (p->http->redirect)
			g_url_free(p->http->redirect);
		if (p->http->proxy)
			g_url_free(p->http->proxy);
		if (p->http->authrealm)
			g_free(p->http->authrealm);
		if (p->http->authpass)
			g_free(p->http->authpass);
		g_free(p->http);
		p->http = NULL;
	}
	if (p->ftp) {
		if (p->ftp->datasocket)
			gnet_tcp_socket_delete(p->ftp->datasocket);
		g_free(p->ftp);
		p->ftp = NULL;
	}
	if (p->file) {
	    g_free(p->file);
	    p->file = NULL;
	}
	g_free(p);
	gt->priv = NULL;

	if (GTK_OBJECT_CLASS(gnome_transfer_parent)->destroy)
		(* GTK_OBJECT_CLASS(gnome_transfer_parent)->destroy) (obj);
}

static void
gnome_transfer_class_init (GnomeTransferClass *klass)
{
	GtkObjectClass *object_class = (GtkObjectClass *) klass;

	object_class->destroy = gnome_transfer_destroy;
	
	gnome_transfer_parent = gtk_type_class (gtk_object_get_type ());

	signals[FILE_INFO] =
		gtk_signal_new ("file_info",
				GTK_RUN_LAST,
				object_class->type,
				GTK_SIGNAL_OFFSET (GnomeTransferClass, file_info),
				gtk_marshal_NONE__POINTER_INT,
				GTK_TYPE_NONE, 2,
				GTK_TYPE_POINTER,
				GTK_TYPE_INT);

	signals[FILE_DATA] =
		gtk_signal_new ("file_data",
				GTK_RUN_LAST,
				object_class->type,
				GTK_SIGNAL_OFFSET (GnomeTransferClass, file_data),
				gtk_marshal_NONE__POINTER_INT,
				GTK_TYPE_NONE, 2,
				GTK_TYPE_POINTER,
				GTK_TYPE_INT);

	signals[FILE_DONE] =
		gtk_signal_new ("file_done",
				GTK_RUN_LAST | GTK_RUN_NO_RECURSE,
				object_class->type,
				GTK_SIGNAL_OFFSET (GnomeTransferClass, file_done),
				gtk_marshal_NONE__INT,
				GTK_TYPE_NONE, 1,
				GTK_TYPE_INT);

	signals[FILE_PAUSED] =
		gtk_signal_new ("file_paused",
				GTK_RUN_LAST,
				object_class->type,
				GTK_SIGNAL_OFFSET (GnomeTransferClass, file_paused),
				gtk_marshal_NONE__INT,
				GTK_TYPE_NONE, 1,
				GTK_TYPE_INT);

	signals[AUTH_REQUIRED] =
		gtk_signal_new ("auth_required",
				GTK_RUN_LAST,
				object_class->type,
				GTK_SIGNAL_OFFSET (GnomeTransferClass, auth_required),
				gtk_marshal_NONE__POINTER,
				GTK_TYPE_NONE, 1,
				GTK_TYPE_POINTER);

	gtk_object_class_add_signals (object_class, signals, LAST_SIGNAL);
}

static void
set_error(GnomeTransfer *t, char *error)
{
	if (t->reason)
		g_free(t->reason);

	if (error)
		t->reason = g_strdup(error);
	else
		t->reason = NULL;
}

static void
gnome_transfer_init (GnomeTransfer *t)
{
	t->priv = g_new0(GnomeTransferPrivate, 1);
	t->headers = g_hash_table_new(g_str_hash, insensitive_equal);
	t->priv->size = -1;
	t->priv->http = g_new0(GnomeTransferHTTP, 1);
	t->priv->ftp = g_new0(GnomeTransferFTP, 1);
	t->priv->file = g_new0(GnomeTransferFile, 1);
}

/**
 * gnome_transfer_new:
 *
 * Create a new GnomeTransfer object.
 * 
 * Return value: A new GnomeTransfer widget.
 **/
GnomeTransfer *
gnome_transfer_new (void)
{
	GnomeTransfer *new = GNOME_TRANSFER(
		gtk_type_new(gnome_transfer_get_type ()));

	new->status = GNOME_TRANSFER_IDLE;

	return new;
}

static char *base64_alphabet =
"ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";

static unsigned char base64_rank[256];
static int base64_init_done = 0;

static int read_buf(char *out, int len, char *in, char *inend)
{
	int c = 0;
	while (c < len && in < inend) {
		*out++ = *in++;
		c++;
	}
	return c;
}

/* must be multiple of 3 */
#define BUFLEN 57

/*

   aaavvvg00
          ^
        ^
         ^

    aaavvvgg0
          ^
         ^
*/

static void
base64_init(void)
{
	int i;

	if (!base64_init_done) {
		base64_init_done=1;
		memset(base64_rank, 0xff, sizeof(base64_rank));
		for (i=0;i<sizeof(base64_rank)/sizeof(base64_rank[0]);i++) {
			base64_rank[(int) base64_alphabet[i]] = i;
		}
		base64_rank['='] = 0;
	}
}

static int
base64_encode(unsigned char *in, int len, unsigned char *out)
{
	unsigned char *realinend = in+len, *inend;
	int count;
	register unsigned char *inptr, *outptr, c1, c2, c3;
	char buf[BUFLEN];

	base64_init();

	inptr = in;
	outptr = out;
	/* assumes read_buf fills our buffer until eof */
	while ( (count = read_buf(buf, BUFLEN, in, realinend)) > 0) {
		inptr = buf;
		inend = buf+count-2;
		in += count;
		while (inptr < inend) {
			c1 = *inptr++;
			c2 = *inptr++;
			c3 = *inptr++;
			*outptr++ = base64_alphabet [ c1 >> 2 ];
			*outptr++ = base64_alphabet [ c2 >> 4 | ( (c1&0x3) << 4 ) ];
			*outptr++ = base64_alphabet [ ( (c2 &0x0f) << 2 ) | (c3 >> 6) ];
			*outptr++ = base64_alphabet [ c3 & 0x3f ];
		}
		switch (inptr-inend) {
		case 0:
			c1 = *inptr++;
			c2 = *inptr++;
			outptr[2] = base64_alphabet [ ( (c2 &0x0f) << 2 ) ];
			goto skip;
		case 1:
			c1 = *inptr++;
			c2=0;
			outptr[2] = '=';
		skip:
			outptr[0] = base64_alphabet [ c1 >> 2 ];
			outptr[1] = base64_alphabet [ c2 >> 4 | ( (c1&0x3) << 4 )];
			outptr[3] = '=';
			outptr += 4;
			break;
		}
/*		*outptr++ = '\n';*/
	}
	*outptr = 0;
	return outptr-out;
}

typedef struct {
    char *name;
    int number;
} MonthTable;

static int
parse_month(char *month)
{
    MonthTable mt[] = {
	{ "Jan", 0 }, { "Feb", 1 }, { "Mar", 2 }, { "Apr", 3 }, { "May", 4 },
	{ "Jun", 5 }, { "Jul", 6 }, { "Aug", 7 }, { "Sep", 8 }, { "Oct", 9 },
	{ "Nov", 10 }, { "Dec", 11 }
    };

    int i;

    for (i = 0; i < 12; i++) {
	if (g_strncasecmp(month, mt[i].name, 3) == 0)
	    return mt[i].number;
    }

    g_assert_not_reached();
    return 0;
} /* parse_month */

static time_t
parse_http_time(char *header)
{
    struct tm *tm;
    time_t tt;
    char **split;
    char **time;
    char **date;

    g_return_val_if_fail(header, -1);

    /* According to RFC 2616, all HTTP 1.1 compliant clients must be able to
       parse dates in any one of the following three formats:

       Sun, 06 Nov 1994 08:49:37 GMT  ; RFC 822, updated by RFC 1123
       Sunday, 06-Nov-94 08:49:37 GMT ; RFC 850, obsoleted by RFC 1036 
       Sun Nov  6 08:49:37 1994       ; ANSI C's asctime() format

       The key, as you can see, is the 4th character, which will tell us
       which format we are in. */

    tm = g_new0(struct tm, 1);

    if (header[3] == ',') {
        /* Sun, 06 Nov 1994 08:49:37 GMT */

        split = g_strsplit(header, " ", 0);
        tm->tm_mday = atoi(split[1]);
        tm->tm_mon = parse_month(split[2]);
        tm->tm_year = atoi(split[3]) - 1900;
        time = g_strsplit(split[4], ":", 0);
        tm->tm_hour = atoi(time[0]);
        tm->tm_min = atoi(time[1]);
        tm->tm_sec = atoi(time[2]);
        g_strfreev(time);
        g_strfreev(split);
    }
    else if (header[3] == ' ') {
        /* Sun Nov  6 08:49:37 1994 */
        int i;

        split = g_strsplit(header, " ", 0);
        tm->tm_mon = parse_month(split[1]);
        i = 2;
        if (!split[i][0])
            i++;
        tm->tm_mday = atoi(split[i++]);
        time = g_strsplit(split[i++], ":", 0);
        tm->tm_hour = atoi(time[0]);
        tm->tm_min = atoi(time[1]);
        tm->tm_sec = atoi(time[2]);
        g_strfreev(time);
        tm->tm_year = atoi(split[i]);
        g_strfreev(split);
    }
    else {
        /* Sunday, 06-Nov-94 08:49:37 GMT */
        int y;

        split = g_strsplit(header, " ", 0);
        date = g_strsplit(split[1], "-", 0);
        tm->tm_mday = atoi(date[0]);
        tm->tm_mon = parse_month(date[1]);
        /* Two digit dates. Wonderful. Those idiots. Lets assume that anything
           before the epoch is a 21st century date and afterward is 20th. */
        y = atoi(date[2]);
        if (y >= 70)
            tm->tm_year = y;
        else
            tm->tm_year = 100 + y;
        g_strfreev(date);
        time = g_strsplit(split[2], ":", 0);
        tm->tm_hour = atoi(time[0]);
        tm->tm_min = atoi(time[1]);
        tm->tm_sec = atoi(time[2]);
        g_strfreev(time);
        g_strfreev(split);
    }

    tm->tm_isdst = -1;
        
    tt = mktime(tm);
    g_free(tm);
    
    return tt;
} /* parse_http_time */

static void
parse_http_statuscode(GnomeTransferHTTP *http, const char *statuscode)
{
	sscanf(statuscode, "HTTP/%*d.%*d %d ", &http->statuscode);
	h(printf(
		"HTTP statuscode %d (%s)\n\n", 
		http->statuscode, statuscode));
	/* Since there can be only one of these per transfer, it's safe to
	   put us into header mode after this. */
	http->state = HTTP_HEADER;
} /* parse_http_statuscode */

static void
parse_http_header(GnomeTransfer *t, char *header)
{
	GnomeTransferPrivate *p = _PRIVATE(t);
	char **split;
	static int content_length = 0, content_range = 0;

	split = g_strsplit(header, ":", 1);

	h(printf("got header %s: %s\n", split[0], split[1]));

	/* Strip the whitespace off the newly split strings. */
	split[0] = g_strstrip(split[0]);
	split[1] = g_strstrip(split[1]);

	g_hash_table_insert(
		t->headers, g_strdup(split[0]), g_strdup(split[1]));

	if (g_strcasecmp(split[0], "Content-Length") == 0) {
		content_length = atoi(split[1]);
	}
	else if (g_strcasecmp(split[0], "Content-Range") == 0) {
		sscanf(split[1], "bytes %*d-%*d/%d", &content_range);
	}
	else if (g_strcasecmp(split[0], "Location") == 0) {
		p->http->redirect = g_url_new(split[1]);
                if (p->http->redirect->port == -1)
                        p->http->redirect->port = 80;
        }
	else if (g_strcasecmp(split[0], "Proxy-Authenticate") == 0) {
		if (p->http->authrealm)
			g_free(p->http->authrealm);
		p->http->authrealm = g_strdup(split[1]);
	}
	else if (g_strcasecmp(split[0], "Last-Modified") == 0) {
	        t->last_modified = parse_http_time(split[1]);
	}

	if (content_range) {
		h(printf("content range\n"));
		p->size = content_range;
		content_range = 0;
	}
	else if (content_length) {
		h(printf("content length\n"));
		p->size = content_length;
		content_length = 0;
	}

	g_strfreev(split);
} /* parse_http_header */

/* The return value from this function determines whether http_read should
   continue with parsing the data */
static gboolean
handle_http_statuscode(GnomeTransfer *t)
{
        GnomeTransferPrivate *p = _PRIVATE(t);

	if (p->http->statuscode == 206 || p->http->statuscode == 200) {
		/* Partial Content or OK */
		p->http->state = HTTP_DATA;
		if (!p->info_emitted) {
			/* We only want to emit this signal for new transfers,
			   not resumed ones */
			gtk_signal_emit(
				GTK_OBJECT(t), signals[FILE_INFO], 
				t->url->path, p->size);
			p->info_emitted = TRUE;
		}
		return TRUE;
	}
	else if (p->http->statuscode == 407 && p->http->authrealm) {
		/* Proxy Authentication Required */
		h(printf("proxy auth required\n"));
		close_transfer(t);
		gtk_signal_emit(
			GTK_OBJECT(t), signals[AUTH_REQUIRED],
			p->http->authrealm);
		return FALSE;
	}
	else if ((p->http->statuscode == 301 ||
                  p->http->statuscode == 302) && p->http->redirect) {
		/* Moved Permanently or Moved Temporarily (1.0)/Found (1.1) */
		char *url;

		close_transfer(t);
		url = g_url_to_string(p->http->redirect, TRUE);
		gnome_transfer_begin_offset(t, url, 0);
		g_free(url);
		return FALSE;
	}
	else {
		/* Another, unhandled, status. Abort. */
		close_transfer(t);
		gtk_signal_emit(
			GTK_OBJECT(t), signals[FILE_DONE], 
			GNOME_TRANSFER_FAILED);
		return FALSE;
	}
} /* parse_http_status */

/* Returning FALSE from this function will mean that this function should
   NOT be called when more data is received. */
static gboolean
http_read(GIOChannel *channel, GIOCondition cond, GnomeTransfer *t)
{
	int bytes;
	GIOError err;
	char buf[BLOCK_SIZE+1];
	char *data = NULL;
	char **split;
	gboolean cont;
	GnomeTransferPrivate *p = _PRIVATE(t);
        int headerbytes = 0;
        int databytes;

	do {
		if (t->status != GNOME_TRANSFER_DATA) {
			h(printf("Unflushed data on the channel. Flushing...\n"));
			g_io_channel_close(channel);
			return FALSE;
		}

		err = g_io_channel_read(channel, buf, BLOCK_SIZE, &bytes);
		
		h(printf("bytes %d, err = %d\n", bytes, err));

		if (p->http->state == HTTP_STATUS || 
		    p->http->state == HTTP_HEADER) {
			int i = 0;

			/* This is kinda hackish, but we need a null character
			   at the end so we can do nice length calculations. */
			buf[bytes] = 0;

			split = g_strsplit(buf, "\r\n", 0);

			/* Step through all of the headers */
			while (split[i] && split[i][0] != 0) {
				if (p->http->state == HTTP_STATUS) {
					parse_http_statuscode(
						p->http, split[i]);
					set_error(t, split[i]);
				}
				else if (p->http->state == HTTP_HEADER)
					parse_http_header(t, split[i]);

                                headerbytes = headerbytes + strlen(split[i]);

				i++;
			}

			cont = handle_http_statuscode(t);
			if (cont == FALSE) {
				/* Don't continue parsing the body; there is
				   an error, we need proxy authorization, or
				   we're being redirected. */
				g_strfreev(split);
				g_io_channel_close(channel);
				return FALSE;
			}

			/* Increment past the null line; everything after this
			   is data */
			i++;

                        /* When we did the split, the \r\n were removed, but
                           we received them in the stream. The HTTP spec
                           requires that they terminate each line, and so we
                           can just add i*2 bytes to headerbytes. */
                        headerbytes = headerbytes + i*2;

                        databytes = bytes - headerbytes;

			/* Check to see if we have data. */
			if (split[i] != NULL)
                                data = g_memdup(buf + headerbytes, databytes);
                        g_strfreev(split);
		}
		else {
                        /* HTTP_DATA */
                        databytes = bytes;
			if (databytes > 0)
                                data = g_memdup(buf, databytes);
		}

		if (t->status == GNOME_TRANSFER_DATA) {
		    if (p->http->state == HTTP_DATA) {
			if (databytes > 0) {
			    p->totalbytes += databytes;
			    gtk_signal_emit(
				    GTK_OBJECT(t), 
				    signals[FILE_DATA],
				    data, databytes);
			    g_free(data);
			}
		    }
		}
		else {
		    /* To ensure that a file_data signal isn't
		       emitted after an aborted/failed transfer */
		    h(printf("Unflushed data coming in on the channel. Flushing...\n"));
		    g_io_channel_close(channel);

		    return FALSE;
		}
	} while (bytes > 0);

	/* G_IO_ERROR_AGAIN means there is more data to read off the channel */
	if (err == G_IO_ERROR_AGAIN)
		return TRUE;

	/* No more bytes to read and no error condition; the file is done */
	if (err == G_IO_ERROR_NONE) {
		g_io_channel_close(channel);
		close_transfer(t);
		gtk_signal_emit(
			GTK_OBJECT(t), signals[FILE_DONE],
			GNOME_TRANSFER_COMPLETE);
		h(printf("transfer complete...\n"));
		return FALSE;
	}
#if 0
	/* An error of some kind occurred on the channel. */
	close_transfer(t);
	gtk_signal_emit(
		GTK_OBJECT(t), signals[FILE_DONE], GNOME_TRANSFER_FAILED);
#endif

	return FALSE;
} /* http_read */

static void
http_connect(GTcpSocket *socket, GTcpSocketNewAsyncStatus status,
	     GnomeTransfer *t)
{
	GIOChannel *channel;
	int bytes;
	GnomeTransferPrivate *p = _PRIVATE(t);

	/* There is a possiblity that the transfer has been freed before this
	   function is called from the gtk main loop */
	if (!p || !p->http)
		return;

	if (t->status != GNOME_TRANSFER_CONNECT) {
		return;
	}

	if (status == GTCP_SOCKET_NEW_ASYNC_STATUS_OK) {
		GString *request;
		char *url;
		char *req;

		/* GNet is a little broken, it seems. */
		{
			int flags;

			flags = fcntl(socket->sockfd, F_GETFL, 0);
			if (flags != -1) {
				fcntl(socket->sockfd, F_SETFL, 
				      flags | O_NONBLOCK);
			}
		}

		url = g_url_to_string(t->url, TRUE);

		g_return_if_fail(t->type != GNOME_TRANSFER_TYPE_NONE);
		if (t->type == GNOME_TRANSFER_TYPE_HEADERS)
			req = "HEAD";
		else
			req = "GET";

		request = g_string_new("");
		g_string_sprintf(
			request, 
			"%s %s HTTP/1.0\r\n"
			"User-Agent: %s\r\n"
                        "Host: %s\r\n",
			req, p->http->proxy ? url : t->url->path,
			p->http->user_agent ? p->http->user_agent : 
			"GnomeTransfer/0.1",
                        t->url->host);
		g_free(url);

		if (p->http->authrealm) {
			h(printf("appending WWW-Authenticate: %s\n", p->http->authrealm));
			g_string_sprintfa(
				request,
				"WWW-Authenticate: %s\r\n", 
				p->http->authrealm);
		}

		if (p->offset > 0) {
			g_string_sprintfa(
				request,
				"Range: bytes=%d-\r\n", p->offset);
		}

		if (p->http->authpass && p->http->proxy) {
			g_string_sprintfa(
				request,
				"Proxy-authorization: Basic %s\r\n",
				p->http->authpass);
		}

		/* End the headers */
		g_string_sprintfa(request, "\r\n");
		
		h(printf("http request: %s\n", request->str));
		channel = gnet_tcp_socket_get_iochannel(socket);
		g_io_channel_write(
			channel, request->str, request->len, &bytes);
		g_string_free(request, TRUE);
		g_io_add_watch_full(channel, G_PRIORITY_LOW, G_IO_IN,
				    (GIOFunc) http_read, t, NULL);
		p->socket = socket;
		p->http->state = HTTP_STATUS;
		t->status = GNOME_TRANSFER_DATA;
	}
	else {
		char *url;
		char *err;
		
		if (p->http->proxy)
			url = g_url_to_string(p->http->proxy, FALSE);
		else
			url = g_url_to_string(t->url, FALSE);
		err = g_strdup_printf(
			"Cannot connect to %s: %s", 
			p->http->proxy ? "proxy" : "server",
			url);
		g_free(url);
		set_error(t, err);
		g_free(err);
		h(printf("Cannot connect to server\n"));
		close_transfer(t);
		gtk_signal_emit(
			GTK_OBJECT(t), signals[FILE_DONE], 
			GNOME_TRANSFER_FAILED);
	}
} /* http_connect */

/* Returning FALSE from this function will mean that this function should
   NOT be called when more data is received. */
static gboolean
ftp_data_read(GIOChannel *channel, GIOCondition cond, GnomeTransfer *t)
{
	int bytes;
	GIOError err;
	char buf[BLOCK_SIZE+1];
	GnomeTransferPrivate *p = _PRIVATE(t);

	do {
		err = g_io_channel_read(channel, buf, BLOCK_SIZE, &bytes);
		
		f(printf("%p: bytes %d, err = %d\n", channel, bytes, err));

		if (t->status == GNOME_TRANSFER_DATA) {
			if (bytes > 0) {
				p->totalbytes += bytes;
				gtk_signal_emit(
					GTK_OBJECT(t), signals[FILE_DATA],
					buf, bytes);
			}
		}
		else {
			/* To ensure that a file_data signal isn't emitted
			   after an aborted/failed/paused transfer */
			f(printf("Unflushed data coming in on the channel\n"));
			return FALSE;
		}
	} while (bytes > 0);

	/* G_IO_ERROR_AGAIN means there is more data to read off the channel */
	if (err == G_IO_ERROR_AGAIN)
		return TRUE;

	/* No more bytes to read and no error condition; the file is done */
	if (bytes == 0 && err == G_IO_ERROR_NONE) {
		close_transfer(t);
		gtk_signal_emit(
			GTK_OBJECT(t), signals[FILE_DONE],
			GNOME_TRANSFER_COMPLETE);
		f(printf("transfer complete...\n"));
		return FALSE;
	}

	/* An error of some kind occurred on the channel. */
	close_transfer(t);
	gtk_signal_emit(
		GTK_OBJECT(t), signals[FILE_DONE], GNOME_TRANSFER_FAILED);
	return FALSE;
} /* ftp_data_read */

static void
ftp_data_connect(GTcpSocket *socket, GTcpSocketNewAsyncStatus status,
		 GnomeTransfer *t)
{
	GIOChannel *channel;
	GnomeTransferPrivate *p = _PRIVATE(t);

	if (status == GTCP_SOCKET_NEW_ASYNC_STATUS_OK) {
		/* GNet is broken */
		{
			int flags;
			flags = fcntl(socket->sockfd, F_GETFL, 0);
			if (flags != -1) {
				fcntl(socket->sockfd, F_SETFL,
				      flags | O_NONBLOCK);
			}
		}

		f(printf("ftp data connection established\n"));

		channel = gnet_tcp_socket_get_iochannel(socket);
		g_io_add_watch_full (channel, G_PRIORITY_LOW, G_IO_IN,
				     (GIOFunc) ftp_data_read, t, NULL);
		p->ftp->datasocket = socket;
		t->status = GNOME_TRANSFER_DATA;
	}
	else {
		char *url;
		char *err;

		url = g_url_to_string(t->url, FALSE);
		err = g_strdup_printf("Cannot connect to host: %s", url);
		set_error(t, err);
		g_free(err);
		g_free(url);
		f(printf("Cannot connect to server\n"));
		close_transfer(t);
		gtk_signal_emit(
			GTK_OBJECT(t), signals[FILE_DONE],
			GNOME_TRANSFER_FAILED);
	}
} /* ftp_data_connect */

static gboolean
ftp_control_read(GIOChannel *channel, GIOCondition cond, GnomeTransfer *t)
{
	int bytes;
	GIOError err;
	char buf[BLOCK_SIZE+1];
	GnomeTransferPrivate *p = _PRIVATE(t);
	int ftpstatus = 0;
	gboolean moreheader = FALSE;
	char *req;
	gboolean cont;
	char *header;
	char *fatal = NULL;

	do {
		char **split;
		int i = 0;

		err = g_io_channel_read(channel, buf, BLOCK_SIZE, &bytes);

		f(printf("reading %d bytes from control channel\n", bytes));

		/* To ensure that control data isn't handled after an
		   aborted/failed/paused transfer */
		if (t->status != GNOME_TRANSFER_DATA) {
			f(printf("unflushed data coming in over the control channel.\n"));
			return FALSE;
		}

		if (!bytes)
			return TRUE;

		buf[bytes] = 0;

		split = g_strsplit(buf, "\n", 0);
		while (split[i] != NULL)
			i++;

		/* Get the last header, complete or not */
		i--;

		/* Incomplete header; we have more to read. */
		if (split[i] && strlen(split[i]) < 4) {
			moreheader = TRUE;
			continue;
		}

		/* If the 4th character is a hyphen, then there is more to
		   read on the control channel.
		   Example: 221-You have transferred blah blah blah
		            221-Total traffic blah blah blah
			    221 Goodbye.
		*/
		if (split[i][3] == '-') {
			moreheader = TRUE;
			continue;
		}

		header = g_strdup(split[i]);

		/* hackish */
		split[i][3] = 0;
		ftpstatus = atoi(split[i]);
		split[i][3] = ' ';
		moreheader = FALSE;
	} while (moreheader);

	cont = FALSE;

	switch (p->ftp->state) {
	case FTP_CONNECT:
		if (ftpstatus/100 == 2) {
			req = "USER anonymous\r\n";
			g_io_channel_write(channel, req, strlen(req), &bytes);
			p->ftp->state = FTP_USER;
			cont = TRUE;
		}
		else {
			f(printf("error during connect.\n"));
			fatal = g_strdup(header);
		}
		break;
	case FTP_USER:
		if (ftpstatus/100 == 2 || ftpstatus/100 == 3) {
			req = g_strdup_printf(
				"PASS %s\r\n", 
				p->ftp->anon_password ? p->ftp->anon_password :
				"transfer@dummy.gnome.org");
			g_io_channel_write(channel, req, strlen(req), &bytes);
			g_free(req);
			req = "PASS\r\n";
			p->ftp->state = FTP_PASS;
			cont = TRUE;
		}
		else {
			f(printf("error during login (username).\n"));
			fatal = g_strdup(header);
		}
		break;
	case FTP_PASS:
		if (ftpstatus/100 == 2 || ftpstatus/100 == 3) {
			req = "TYPE I\r\n";
			g_io_channel_write(channel, req, strlen(req), &bytes);
			p->ftp->state = FTP_TYPE;
			cont = TRUE;
		}
		else {
			f(printf("error during login (password)\n"));
			fatal = g_strdup(header);
		}
		break;
	case FTP_TYPE:
		if (ftpstatus/100 == 2 || ftpstatus/100 == 3) {
			req = g_strdup_printf("SIZE %s\r\n", t->url->path);
			g_io_channel_write(channel, req, strlen(req), &bytes);
			g_free(req);
			req = "SIZE\r\n";
			p->ftp->state = FTP_SIZE;
			cont = TRUE;
		}
		else {
			f(printf("error setting type\n"));
			fatal = g_strdup(header);
		}
		break;
	case FTP_SIZE:
		if (ftpstatus == 213) {
			/* By adding 4, we're skipping the status code and
			   the space, so we can get the actual size. 
			   Example: 213 42748
			*/

			p->size = atoi(header+4);

			if (!p->size)
				p->size = -1;
		}
		else
			p->size = -1;
		f(printf("Header: %s\nSize: %d\n", header, p->size));
		req = "PASV\r\n";
		g_io_channel_write(channel, req, strlen(req), &bytes);
		p->ftp->state = FTP_PASV;
		cont = TRUE;
		break;
	case FTP_PASV:
		if (ftpstatus/100 == 2 || ftpstatus/100 == 3) {
			char *b;
			int hi, lo;
			int q1, q2, q3, q4;
			char *addrtext;
			GInetAddr *addr;
			
			b = strchr(header, '(');
			/* 227 Entering Passive Mode (206,166,189,128,20,6) */
			sscanf(b + 1, "%d,%d,%d,%d,%d,%d", 
			       &q1, &q2, &q3, &q4, &hi, &lo);
			addrtext = g_strdup_printf(
				"%d.%d.%d.%d", q1, q2, q3, q4);
			f(printf("passive listening on %s:%d\n",
				 addrtext, hi*256+lo));
			addr = gnet_inetaddr_new(addrtext, hi*256+lo);
			g_free(addrtext);
			p->ftp->dataid = gnet_tcp_socket_new_async(
				addr, 
				(GTcpSocketNewAsyncFunc) ftp_data_connect, t);
			gnet_inetaddr_unref(addr);

			if (p->offset > 0) {
				req = g_strdup_printf(
					"REST %d\r\n", p->offset);
				p->ftp->state = FTP_REST;
			}
			else {
				req = g_strdup_printf(
					"RETR %s\r\n", t->url->path);
				p->ftp->state = FTP_RETR;
			}
			g_io_channel_write(channel, req, strlen(req), &bytes);
			g_free(req);
			req = (p->offset > 0) ? "REST\r\n" : "RETR\r\n";
			cont = TRUE;
		}
		else {
			f(printf("error switching into passive mode\n"));
			fatal = g_strdup(header);
		}
		break;
	case FTP_REST:
		if (ftpstatus/100 != 3) {
			f(printf("REST unimplemented server-side. Restarting from beginning.\n"));
			p->offset = 0;
		}
		req = g_strdup_printf("RETR %s\r\n", t->url->path);
		p->ftp->state = FTP_RETR;
		g_io_channel_write(channel, req, strlen(req), &bytes);
		f(printf("retrieving: %s\n", req));
		g_free(req);
		req = "RETR\r\n";
		cont = TRUE;
		break;
	case FTP_RETR:
		if (ftpstatus/100 == 1) {
			if (!p->info_emitted) {
				gtk_signal_emit(
					GTK_OBJECT(t), signals[FILE_INFO],
					t->url->path, p->size);
				p->info_emitted = TRUE;
			}
			f(printf("retreiving data\n"));
			req = "(retrieving data)\n";
			cont = TRUE;
		}
		else if (ftpstatus/100 == 2 || ftpstatus/100 == 3) {
			f(printf("data transfer complete\n"));
			req = "QUIT\r\n";
			g_io_channel_write(channel, req, strlen(req), &bytes);
			p->ftp->state = FTP_QUIT;
			cont = TRUE;
		}
		else {
			f(printf("Unable to transfer file\n"));
			fatal = g_strdup(header);
		}
		break;
	case FTP_QUIT:
		close_transfer(t);
		break;
	default:
		f(printf("unknown state. Bail!\n"));
		req = "QUIT\r\n";
		g_io_channel_write(channel, req, strlen(req), &bytes);
		p->ftp->state = FTP_QUIT;
		fatal = g_strdup("Protocol failure");
		break;
	}

	f(printf("Request: %s", req));

	if (fatal) {
		g_warning("Fatal transfer error: %s\n", fatal);
		set_error(t, fatal);
		g_free(fatal);
		gtk_signal_emit(
			GTK_OBJECT(t), signals[FILE_DONE],
			GNOME_TRANSFER_FAILED);
		close_transfer(t);
	}

	g_free(header);
	return cont;
} /* ftp_control_read */

static void
ftp_control_connect(GTcpSocket *socket, GTcpSocketNewAsyncStatus status,
		    GnomeTransfer *t)
{
	GIOChannel *channel;
	GnomeTransferPrivate *p = _PRIVATE(t);

	if (status == GTCP_SOCKET_NEW_ASYNC_STATUS_OK) {
		f(printf("connected to ftp control channel\n"));
		
		p->ftp->state = FTP_CONNECT;

		/* Have I mentioned that GNet is broken? */
		{
			int flags;
			flags = fcntl(socket->sockfd, F_GETFL, 0);
			if (flags != -1)
				fcntl(socket->sockfd, F_SETFL,
				      flags | O_NONBLOCK);
		}

		channel = gnet_tcp_socket_get_iochannel(socket);
		g_io_add_watch_full (channel, G_PRIORITY_LOW, G_IO_IN,
				     (GIOFunc) ftp_control_read, t, NULL);
		p->socket = socket;
		t->status = GNOME_TRANSFER_DATA;
	}
	else {
		char *url;
		char *err;

		url = g_url_to_string(t->url, FALSE);
		err = g_strdup_printf(
			"Cannot connect to control channel: %s", url);
		set_error(t, err);
		g_free(err);
		g_free(url);
		f(printf("Cannot connect to control channel\n"));
		close_transfer(t);
		gtk_signal_emit(
			GTK_OBJECT(t), signals[FILE_DONE],
			GNOME_TRANSFER_FAILED);
	}
} /* ftp_control_connect */

static gboolean
file_read(GIOChannel *channel, GIOCondition cond, GnomeTransfer *t)
{
    int bytes;
    GIOError err;
    char buf[BLOCK_SIZE+1];
    GnomeTransferPrivate *p = _PRIVATE(t);

    err = g_io_channel_read(channel, buf, BLOCK_SIZE, &bytes);
    
    p->totalbytes += bytes;
    if (bytes) {
	    gtk_signal_emit(GTK_OBJECT(t), signals[FILE_DATA], buf, bytes);
    }

    /* G_IO_ERROR_AGAIN means there is more data to read off the channel */
    if (err == G_IO_ERROR_AGAIN)
	    return TRUE;
    
    /* There is more data to read */
    if (bytes > 0 && err == G_IO_ERROR_NONE)
	    return TRUE;
    
    /* No more bytes to read and no error condition; the file is done */
    if (err == G_IO_ERROR_NONE) {
	    g_io_channel_unref(channel);
	    close_transfer(t);
	    gtk_signal_emit(
		    GTK_OBJECT(t), signals[FILE_DONE],
		    GNOME_TRANSFER_COMPLETE);
	    return FALSE;
    }

    /* An error of some kind occurred on the channel */
    close_transfer(t);
    gtk_signal_emit(GTK_OBJECT(t), signals[FILE_DONE], GNOME_TRANSFER_FAILED);
    return FALSE;
} /* file_data */

static void
gotaddress(GInetAddr *addr, GInetAddrAsyncStatus status, char *name, GnomeTransfer *t)
{
	struct _GnomeTransferPrivate *p = _PRIVATE(t);

	if (t->status == GNOME_TRANSFER_ABORTED)
		return;

	if (status == GINETADDR_ASYNC_STATUS_OK) {
		d(printf("connecting to %s\n", name));

		p->dataid = gnet_tcp_socket_new_async(
			addr,
			(GTcpSocketNewAsyncFunc) p->connect_func,
			t);
		t->status = GNOME_TRANSFER_CONNECT;
	} else {
		d(printf("could not resolve '%s'...\n", name));
		t->status = GNOME_TRANSFER_FAILED;
		gtk_signal_emit((GtkObject *)t, signals[FILE_DONE], GNOME_TRANSFER_FAILED);
	}
}

static int
begin(GnomeTransfer *t)
{
	GTcpSocketNewAsyncFunc connectfunc;
	GnomeTransferPrivate *p = _PRIVATE(t);
	GInetAddr *addr;
	Gurl *gu;

	t->type = GNOME_TRANSFER_TYPE_DATA;

	if (g_strcasecmp(t->url->protocol, "file") == 0) {
		int fd;
		struct stat fdstat;
		
		fd = open(t->url->path, O_RDONLY);
		if (fd < 0)
			return -1;
		if (fstat(fd, &fdstat))
			return -1;
		t->status = GNOME_TRANSFER_DATA;
		if (!p->info_emitted) {
			gtk_signal_emit(
				GTK_OBJECT(t), signals[FILE_INFO], 
				t->url->path, fdstat.st_size);
			p->info_emitted = TRUE;
		}
		if (t->status == GNOME_TRANSFER_DATA) {
			p->file->iochannel = g_io_channel_unix_new(fd);
			g_io_add_watch_full(
				p->file->iochannel, G_PRIORITY_LOW,
				G_IO_IN, (GIOFunc) file_read, t, NULL);
		}
	    return 0;
	}

	if (g_strcasecmp(t->url->protocol, "http") == 0) {
		connectfunc = (GTcpSocketNewAsyncFunc) http_connect;

		if (t->url->port == -1)
			t->url->port = 80;

		if (p->http->proxy) {
			gu = p->http->proxy;
			if (gu->port == -1)
				gu->port = 80;
		}
		else
			gu = t->url;

		p->type = TRANSFER_HTTP;
	} else if (g_strcasecmp(t->url->protocol, "ftp") == 0) {
		connectfunc = (GTcpSocketNewAsyncFunc) ftp_control_connect;
		if (t->url->port == -1)
			t->url->port = 21;
		gu = t->url;
		p->type = TRANSFER_FTP;
	} else {
		t->status = GNOME_TRANSFER_FAILED;
		g_url_free(t->url);
		t->url = NULL;
		return -1;
	}

	d(printf("looking up host\n"));

	t->status = GNOME_TRANSFER_LOOKUP;
	p->connect_func = connectfunc;
	addr = gnet_inetaddr_new(gu->host, gu->port);

	if (!addr) {
		char *err;

		if (p->type == TRANSFER_HTTP && p->http->proxy)
			err = g_strdup_printf(
				"Could not resolve proxy: %s", gu->host);
		else
			err = g_strdup_printf(
				"Could not resolve host: %s", gu->host);
		set_error(t, err);
		g_free(err);
		t->status = GNOME_TRANSFER_FAILED;
		return -1;
	}

	gotaddress(addr, GINETADDR_ASYNC_STATUS_OK, gu->host, t);
	gnet_inetaddr_unref(addr);
	t->status = GNOME_TRANSFER_CONNECT;

	return 0;
}

static void
gnome_transfer_reset(GnomeTransfer *t)
{
	GnomeTransferPrivate *p = _PRIVATE(t);

	if (t->url) {
		g_url_free(t->url);
		t->url = NULL;
	}

	if (t->reason) {
		g_free(t->reason);
		t->reason = NULL;
	}

	g_hash_table_foreach(t->headers, hash_free, NULL);
	g_hash_table_destroy(t->headers);
	t->headers = g_hash_table_new(g_str_hash, insensitive_equal);

	p->type = TRANSFER_NONE;
	p->connect_func = NULL;
	p->dataid = NULL;
	p->size = -1;
	p->totalbytes = 0;
	p->offset = 0;
	p->info_emitted = FALSE;

	p->http->statuscode = 0;

	if (p->http->redirect) {
		g_url_free(p->http->redirect);
		p->http->redirect = NULL;
	}

	p->ftp->dataid = NULL;
	p->ftp->datasocket = NULL;
} /* gnome_transfer_reset */

int
gnome_transfer_begin_offset(GnomeTransfer *t, char *uri, int offset)
{
	GnomeTransferPrivate *p = _PRIVATE(t);
	gboolean info_emitted = p->info_emitted;

	gnome_transfer_reset(t);
	/* HACK! HACK! HACK! */
	p->info_emitted = info_emitted;

	t->status = GNOME_TRANSFER_IDLE;

	t->url = g_url_new(uri);

	if (t->url == NULL) {
		t->status = GNOME_TRANSFER_FAILED;
		set_error(t, "Invalid URL");
		return -1;
	}

	if (t->url->protocol == NULL) {
		t->status = GNOME_TRANSFER_FAILED;
		set_error(t, "Invalid URL, no protocol");
		g_url_free(t->url);
		t->url = NULL;
		return -1;
	}

	p->offset = offset;

	return begin(t);
}
 
int
gnome_transfer_begin(GnomeTransfer *t, char *uri)
{
	return gnome_transfer_begin_offset(t, uri, 0);
} /* gnome_transfer_begin */

int
gnome_transfer_retry(GnomeTransfer *t)
{
	char *url;
	int res;

	url = g_url_to_string(t->url, TRUE);
	res = gnome_transfer_begin(t, url);
	g_free(url);

	return res;
}

int
gnome_transfer_headers(GnomeTransfer *t, char *uri)
{
	GnomeTransferPrivate *p = _PRIVATE(t);
	GTcpSocketNewAsyncFunc connectfunc;
	Gurl *gu;
	GInetAddr *addr;

	t->url = g_url_new(uri);

	if (t->url == NULL) {
		t->status = GNOME_TRANSFER_FAILED;
		set_error(t, "Invalid URL");
		return -1;
	}

	if (t->url->protocol == NULL) {
		t->status = GNOME_TRANSFER_FAILED;
		set_error(t, "Invalid URL, no protocol");
		g_url_free(t->url);
		t->url = NULL;
		return -1;
	}

	t->type = GNOME_TRANSFER_TYPE_HEADERS;

	if (g_strcasecmp(t->url->protocol, "file") == 0)
		return -1;
	else if (g_strcasecmp(t->url->protocol, "ftp") == 0) {
		/* FIXME: Write me. */
		return -1;
	}
	else if (g_strcasecmp(t->url->protocol, "http") == 0) {
		connectfunc = (GTcpSocketNewAsyncFunc) http_connect;

		if (t->url->port == -1)
			t->url->port = 80;

		if (p->http->proxy) {
			gu = p->http->proxy;
			if (gu->port == -1 )
				gu->port = 80;
		}
		else
			gu = t->url;

		p->type = TRANSFER_HTTP;
	}

	t->status = GNOME_TRANSFER_LOOKUP;
	p->connect_func = connectfunc;
	addr = gnet_inetaddr_new(gu->host, gu->port);

	if (!addr) {
		char *err;

		err = g_strdup_printf(
			"Could not resolve %s: %s",
			(p->type == TRANSFER_HTTP && p->http->proxy) ?
			"proxy" : "host", gu->host);
		set_error(t, err);
		g_free(err);
		t->status = GNOME_TRANSFER_FAILED;
		return -1;
	}

	gotaddress(addr, GINETADDR_ASYNC_STATUS_OK, gu->host, t);
	gnet_inetaddr_unref(addr);
	t->status = GNOME_TRANSFER_CONNECT;

	return 0;
} /* gnome_transfer_headers */


static void
close_transfer(GnomeTransfer *t)
{
	GnomeTransferPrivate *p = _PRIVATE(t);
	GIOChannel *channel;

	if (!p)
		return;

	if (p->socket) {
		channel = gnet_tcp_socket_get_iochannel(p->socket);
		g_io_channel_close(channel);
		gnet_tcp_socket_delete(p->socket);
		p->socket = NULL;
	}

	if (p->ftp && p->ftp->datasocket) {
		channel = gnet_tcp_socket_get_iochannel(p->ftp->datasocket);
		g_io_channel_close(channel);
		gnet_tcp_socket_delete(p->ftp->datasocket);
		p->ftp->datasocket = NULL;
	}

	if (p->dataid) {
		/* gnet_tcp_socket_new_async_cancel(p->dataid); */
		p->dataid = NULL;
	}

	if (p->ftp && p->ftp->dataid) {
		/* gnet_tcp_socket_new_async_cancel(p->ftp->dataid); */
		p->ftp->dataid = NULL;
	}

	p->info_emitted = FALSE;

	d(printf("transfer closed\n"));
}

void
gnome_transfer_abort(GnomeTransfer *t)
{
	if (t->status != GNOME_TRANSFER_ABORTED &&
	    t->status != GNOME_TRANSFER_COMPLETE &&
	    t->status != GNOME_TRANSFER_FAILED) {
		t->status = GNOME_TRANSFER_ABORTED;
		close_transfer(t);
		gtk_signal_emit(
			GTK_OBJECT(t), signals[FILE_DONE], 
			GNOME_TRANSFER_ABORTED);
	}
}

void
gnome_transfer_pause(GnomeTransfer *t)
{
	GnomeTransferPrivate *p = _PRIVATE(t);

	if (t->status != GNOME_TRANSFER_PAUSED) {
		gboolean info_emitted;

		t->status = GNOME_TRANSFER_PAUSED;
		info_emitted = p->info_emitted;
		close_transfer(t);
		p->info_emitted = info_emitted;
		gtk_signal_emit(
			GTK_OBJECT(t), signals[FILE_PAUSED],
			p->offset + p->totalbytes);
	} else {
		g_warning("Pausing an already paused transfer, pauses do not nest");
	}
}

void
gnome_transfer_resume(GnomeTransfer *t)
{
	struct _GnomeTransferPrivate *p = _PRIVATE(t);

	if (t->status == GNOME_TRANSFER_PAUSED) {
		char *url;

		url = g_url_to_string(t->url, TRUE);
		if (t->type == GNOME_TRANSFER_TYPE_HEADERS) {
			gnome_transfer_headers(t, url);
		}
		else {
			gnome_transfer_begin_offset(
				t, url, p->offset + p->totalbytes);
		}
		
		g_free(url);
	} else {
		g_warning("Trying to resume a non-paused transfer, ignored");
	}
}

int 
gnome_transfer_percent(GnomeTransfer *t)
{
	int bytes;
	struct _GnomeTransferPrivate *p = _PRIVATE(t);

	bytes = p->totalbytes;

	if (p->size == -1) {
		return (bytes % 100000) / 1000; /* make up a changing number?? */
	} else {
		return (bytes * 100)/p->size;
	}
}

int
gnome_transfer_bytes(GnomeTransfer *t)
{
	int bytes;
	struct _GnomeTransferPrivate *p = _PRIVATE(t);

	bytes = p->totalbytes + p->offset;

	return bytes;
}

int
gnome_transfer_total(GnomeTransfer *t)
{
	struct _GnomeTransferPrivate *p = _PRIVATE(t);
	return p->size;
}

char *
gnome_transfer_reason(GnomeTransfer *t)
{
	if (t->reason)
		return t->reason;

	switch (t->status) {
	case GNOME_TRANSFER_IDLE:
	case GNOME_TRANSFER_LOOKUP:
	case GNOME_TRANSFER_CONNECT:
	case GNOME_TRANSFER_DATA:
	case GNOME_TRANSFER_COMPLETE:
		return "";
	default:
		break;
	}
	return "Unknown reason";
}

void 
gnome_transfer_proxy_password(GnomeTransfer *t, char *name, char *pass)
{
	struct _GnomeTransferPrivate *p = _PRIVATE(t);
	char *raw, *enc;

	if (p->http->authpass)
		g_free(p->http->authpass);
	p->http->authpass = NULL;
	if (name && pass) {
		raw = g_strdup_printf(
			"%s:%s", name ? name : "", pass ? pass : "");
		enc = g_malloc(((strlen(raw)*4)/3) + 4);
		base64_encode(raw, strlen(raw), enc);
		p->http->authpass = enc;
		g_free(raw);
	}
}

int 
gnome_transfer_proxy(GnomeTransfer *t, char *uri)
{
	GnomeTransferPrivate *p = _PRIVATE(t);

	/* Not an HTTP transfer? */
	if (!p->http)
		return -1;

	if (p->http->proxy) {
		g_url_free(p->http->proxy);
		p->http->proxy = NULL;
	}

	if (!uri)
		return 0;

	p->http->proxy = g_url_new(uri);
	if (p->http->proxy == NULL)
		return -1;

	if (p->http->proxy->port == -1)
		p->http->proxy->port = 80;

	return 0;
}

static void
verify_file_info(GnomeTransfer *t, char *name, int size, gboolean *success)
{
	gnome_transfer_abort(t);
	*success = TRUE;
	gtk_main_quit();
} /* verify_file_info */

static void
verify_file_done(GnomeTransfer *t, GnomeTransferStatus status, 
		 gboolean *success)
{
	if (status == GNOME_TRANSFER_FAILED) {
		*success = FALSE;
		gtk_main_quit();
	}
} /* verify_file_done */

gboolean
gnome_transfer_verify(char *uri)
{
	GnomeTransfer *t;
	gboolean success;

	t = gnome_transfer_new();

	/* So. Here's the plan: We're connecting to the file_info signal
	   because if the file does exist, it'll give us the info and we don't
	   actually want to download the file. We'll abort the transfer and
	   quit out of out main loop. If the file does -not- exist, we'll get
	   a file_done signal with a GNOME_TRANSFER_FAILED. If that's the case,
	   we don't need to abort the transfer and we can just quit out of
	   our main loop. Simple, hmm?
	*/

	gtk_signal_connect(
		GTK_OBJECT(t), "file_info", verify_file_info, &success);
	gtk_signal_connect(
		GTK_OBJECT(t), "file_done", verify_file_done, &success);
	
	if (gnome_transfer_begin(t, uri))
		return FALSE;

	gtk_main();

	gtk_object_destroy(GTK_OBJECT(t));

	return success;
} /* gnome_transfer_verify */

void
gnome_transfer_emit_info(GnomeTransfer *t, char *name, int size)
{
    gtk_signal_emit(GTK_OBJECT(t), signals[FILE_INFO], name, size);
} /* gnome_transfer_emit_info */

void
gnome_transfer_emit_data(GnomeTransfer *t, char *buf, int size)
{
    gtk_signal_emit(GTK_OBJECT(t), signals[FILE_DATA], buf, size);
} /* gnome_transfer_emit_data */

void
gnome_transfer_emit_done(GnomeTransfer *t, GnomeTransferStatus status)
{
    gtk_signal_emit(GTK_OBJECT(t), signals[FILE_DONE], status);
} /* gnome_transfer_emit_done */
