/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
/*
 * simple.c: Simple IRC Client
 *
 * Authors:
 *      Alex Graveley (alex@ximian.com)
 *
 * Copyright (C) 2001, Ximian, Inc.
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <gtkhtml/gtkhtml.h>

#include <gnome.h>
#include <libgnomeui/gnome-window-icon.h>
#include <gdk-pixbuf/gdk-pixbuf.h>
#include <gdk-pixbuf/gnome-canvas-pixbuf.h>
#include <gconf/gconf-client.h>

#include "../common/fe.h"
#include "../common/xchat.h"
#include "../common/dcc.h"
#include "../common/userlist.h"
#include "../libgnometransfer/gnome-transfer.h"
#include "simple.h"
#include "pixmaps.h"
#include "xtext.h"
#include "gtkutil.h"

extern struct xchatprefs prefs;
extern GSList *sess_list;
extern struct server *new_server (void);
extern struct session *new_session (struct server *serv, char *from);
extern void connect_server (struct session *sess, 
			    char *server, 
			    int port,
			    int quiet);
extern void disconnect_server (struct session *sess, int sendquit, int err);
extern int is_session (session * sess);
extern void key_handle_send (GtkWidget * ignored, struct session *sess);
extern int key_handle_key_press (GtkWidget * wid, 
				 GdkEventKey * evt, 
				 struct session *sess);
extern GtkStyle *inputgad_style;
extern void PrintTextRaw (GtkWidget * textwidget, 
			  unsigned char *text,
			  int indent);
extern void PrintText (struct session *, char *);
extern int kill_session_callback (struct session *sess);
extern char *nocasestrstr (char *text, char *tofind);
extern struct session *find_dialog (struct server *serv, char *nick);
extern void userlist_dnd_drop (GtkWidget * widget, 
			       GdkDragContext * context,
			       gint x, 
			       gint y,
			       GtkSelectionData * selection_data,
			       guint info, 
			       guint32 time, 
			       struct session *sess);
extern int userlist_dnd_motion (GtkWidget * widget, 
				GdkDragContext * context, 
				gint x,
				gint y, 
				guint ttime);
extern int userlist_dnd_leave (GtkWidget * widget, 
			       GdkDragContext * context, 
			       guint ttime);

extern void nick_command_parse (session * sess, char *cmd, char *nick,
				char *allnick);

extern void chanlist_opengui (struct server * sess);

extern void menu_nickmenu (session * sess, GdkEventButton * event, char *nick);
extern void menu_middlemenu (session * sess, GdkEventButton * event);
extern void menu_urlmenu (struct session *sess, GdkEventButton * event,
			  char *url);
extern void menu_hostmenu (struct session *sess, GdkEventButton * event,
			   char *url);
extern void menu_chanmenu (struct session *sess, GdkEventButton * event,
			   char *url);
extern void key_init (void);

static void connect_to_support_channel (struct session *sess);
static gchar *construct_url (gchar *url);

extern session *menu_sess;

static GByteArray *support_server_result = NULL;
static gchar *support_server = NULL;
static gint support_server_port = 6667;
static gchar *support_channel = NULL;
static gchar *support_keyword = NULL;
static gchar *support_greeting = NULL;

GdkColor colors[] = {
	{0, 0x0000, 0x0000, 0x0000},	/* 0  black */
	{0, 0xcf3c, 0xcf3c, 0xcf3c},	/* 1  white */
	{0, 0x0000, 0x0000, 0xcccc},	/* 2  blue */
	{0, 0x0000, 0xcccc, 0x0000},	/* 3  green */
	{0, 0xcccc, 0x0000, 0x0000},	/* 4  red */
	{0, 0xbbbb, 0xbbbb, 0x0000},	/* 5  yellow/brown */
	{0, 0xbbbb, 0x0000, 0xbbbb},	/* 6  purple */
	{0, 0xffff, 0xaaaa, 0x0000},	/* 7  orange */
	{0, 0xeeee, 0xdddd, 0x2222},	/* 8  yellow */
	{0, 0x3333, 0xdede, 0x5555},	/* 9  green */
	{0, 0x0000, 0xcccc, 0xcccc},	/* 10 aqua */
	{0, 0x3333, 0xdddd, 0xeeee},	/* 11 light aqua */
	{0, 0x0000, 0x0000, 0xffff},	/* 12 blue */
	{0, 0x9999, 0x2222, 0xeeee},	/* 13 light purple */
	{0, 0x7777, 0x7777, 0x7777},	/* 14 grey */
	{0, 0x9999, 0x9999, 0x9999},	/* 15 light grey */

	{0, 0x0000, 0x0000, 0x8b8b},	/* 16 marktext Back (blue) */
	{0, 0xd6d6, 0xd6d6, 0xd6d6},	/* 17 marktext Fore (white) */
	{0, 0x0000, 0x0000, 0x0000},	/* 18 foreground (black) */
	{0, 0xffff, 0xffff, 0xffff},	/* 19 background (white) */
};

static void 
palette_alloc (GtkWidget * widget)
{
	int i;

	if (!colors[0].pixel) {	/* don't do it again */
		for (i = 0; i < 20; i++) {
			colors[i].pixel =
			    (gulong) ((colors[i].red & 0xff00) * 256 +
				      (colors[i].green & 0xff00) +
				      (colors[i].blue & 0xff00) / 256);
			if (!gdk_color_alloc
			    (gtk_widget_get_colormap (widget), &colors[i]))
				fprintf (stderr,
					 "*** X-CHAT: cannot alloc colors\n");
		}
	}
}

#define WORD_URL     1
#define WORD_NICK    2
#define WORD_CHANNEL 3
#define WORD_HOST    4
#define WORD_EMAIL   5
#define WORD_DIALOG  -1

int
simple_check_word_url (GtkXText * xtext, char *word)
{
	if (!strncasecmp (word, "http:", 5)) return WORD_URL;
	if (!strncasecmp (word, "https:", 6)) return WORD_URL;
	if (!strncasecmp (word, "www.", 4)) return WORD_URL;
	if (!strncasecmp (word, "ftp.", 4)) return WORD_URL;
	if (!strncasecmp (word, "ftp:", 4)) return WORD_URL;

	return 0;
}

/* check if a word is clickable */

static int 
maingui_word_check (GtkXText * xtext, char *word)
{
	session *sess;
	char *at, *dot;
	int i, dots;
	int len = strlen (word);

	if (!strncasecmp (word, "http:", 5)) return WORD_URL;
	if (!strncasecmp (word, "https:", 6)) return WORD_URL;
	if (!strncasecmp (word, "www.", 4)) return WORD_URL;
	if (!strncasecmp (word, "ftp.", 4)) return WORD_URL;
	if (!strncasecmp (word, "ftp:", 4)) return WORD_URL;
	if (!strncasecmp (word, "irc://", 6)) return WORD_URL;
	if (!strncasecmp (word, "irc.", 4)) return WORD_URL;

	if ((word[0] == '@' || word[0] == '+') && word[1] == '#')
		return WORD_CHANNEL;

	if (word[0] == '#' && word[1] != '#' && word[1] != 0)
		return WORD_CHANNEL;

	sess = gtk_object_get_user_data (GTK_OBJECT (xtext));

	if (find_name (sess, word)) return WORD_NICK;

	at = strchr (word, '@');	/* check for email addy */
	dot = strrchr (word, '.');
	if (at && 
	    dot &&
	    (unsigned long) at < (unsigned long) dot) {
			if (strchr (word, '*')) return WORD_HOST;
			else return WORD_EMAIL;
	}

	/* check if it's an IP number */
	dots = 0;
	for (i = 0; i < len; i++) {
		if (word[i] == '.') dots++;
	}

	if (dots == 3 && inet_addr (word) != -1) return WORD_HOST;

	if (!strncasecmp (word + len - 5, ".html", 5)) return WORD_HOST;
	if (!strncasecmp (word + len - 4, ".org", 4)) return WORD_HOST;
	if (!strncasecmp (word + len - 4, ".net", 4)) return WORD_HOST;
	if (!strncasecmp (word + len - 4, ".com", 4)) return WORD_HOST;
	if (!strncasecmp (word + len - 4, ".edu", 4)) return WORD_HOST;

	if (len > 5 &&
	    word[len - 3] == '.' &&
	    isalpha (word[len - 2]) && 
	    isalpha (word[len - 1])) return WORD_HOST;

	if (sess->is_dialog) return WORD_DIALOG;

	return 0;
}

static void
maingui_word_clicked (GtkXText * xtext, 
		      char *word, 
		      GdkEventButton * even,
		      session * sess)
{
	if (even->button == 1) {	/* left button */
		if (even->state & GDK_CONTROL_MASK) {
			switch (maingui_word_check (xtext, word)) {
			case WORD_URL:
			case WORD_HOST:
				gnome_url_show (word);
			}
		}
		return;
	}

	if (even->button == 2) {
		if (sess->is_dialog)
			menu_middlemenu (sess, even);
		return;
	}

	switch (maingui_word_check (xtext, word)) {
	case 0:
		menu_middlemenu (sess, even);
		break;
	case WORD_URL:
	case WORD_HOST:
		menu_urlmenu (sess, even, word);
		break;
	case WORD_NICK:
		menu_nickmenu (sess, even, word);
		break;
	case WORD_CHANNEL:
		if (*word == '@' || *word == '+')
			word++;
		menu_chanmenu (sess, even, word);
		break;
	case WORD_EMAIL:
		{
			char *newword = malloc (strlen (word) + 10);
			if (*word == '~')
				word++;
			sprintf (newword, "mailto:%s", word);
			menu_urlmenu (sess, even, newword);
			free (newword);
		}
		break;
	case WORD_DIALOG:
		menu_nickmenu (sess, even, sess->channel);
		break;
	}
}

void 
fe_userlist_hide (struct session *sess)
{
	if (sess->userlisthidden) {
		gtk_widget_show (sess->gui->userlistbox);
		sess->userlisthidden = FALSE;
		prefs.hideuserlist = 0;
	} else {
		sess->userlisthidden = TRUE;
		gtk_widget_hide (sess->gui->userlistbox);
		prefs.hideuserlist = 1;
	}
}

static void
handle_tooltip (GdkEvent *ev, struct session *sess, gchar *tip)
{
	if (ev->type == GDK_ENTER_NOTIFY) {
		gint ret;

		gtk_tooltips_set_tip (sess->gui->tooltip, 
				      GTK_WIDGET (sess->gui->title_canvas), 
				      tip,
				      NULL);

		gtk_tooltips_enable (sess->gui->tooltip);

		gtk_signal_emit_by_name (GTK_OBJECT (sess->gui->title_canvas),
					 "event",
					 ev,
					 &ret);
	} else if (ev->type == GDK_LEAVE_NOTIFY)
		gtk_tooltips_disable (sess->gui->tooltip);
}

static int
click_userlist (GtkWidget *w, GdkEvent *ev, struct session *sess)
{
	if (GTK_WIDGET_VISIBLE (sess->gui->userlistbox)) {
		handle_tooltip (ev, sess, "Hide Channel User List");
	} else {
		handle_tooltip (ev, sess, "Show Channel User List");
	}

	if (ev->type != GDK_BUTTON_RELEASE) return 0;

	fe_userlist_hide (sess);
	return 1;
}

static int
click_chanlist (GtkWidget *w, GdkEvent *ev, struct session *sess)
{
	handle_tooltip (ev, sess, "Go to a Different Support Channel");

	if (ev->type != GDK_BUTTON_RELEASE) return 0;

	chanlist_opengui (sess->server);

	return 1;
}

static void
title_configure_size (GtkWidget *w, GtkAllocation *alloc, struct session *sess)
{
        gnome_canvas_set_scroll_region (GNOME_CANVAS (w), 0.0, 0.0,
                                        alloc->width, alloc->height);

        gnome_canvas_item_set (sess->gui->title_box,
                               "x1", 0.0,
                               "y1", 0.0,
                               "x2", (gfloat) alloc->width,
                               "y2", (gfloat) alloc->height,
                               "width_units", 1.0, 
			       NULL);

        gnome_canvas_item_set (sess->gui->banner,
                               "x", 15.0,
                               "y", 24.0,
                               "anchor", GTK_ANCHOR_WEST, 
			       NULL);

        gnome_canvas_item_set (sess->gui->chanlist_btn,
                               "x", (gfloat) (alloc->width - 88),
                               "y", (gfloat) ((alloc->height / 2) - 16),
                               NULL);

        gnome_canvas_item_set (sess->gui->userlist_btn,
                               "x", (gfloat) (alloc->width - 44),
                               "y", (gfloat) ((alloc->height / 2) - 16),
                               NULL);
}

static void
side_configure_size (GtkWidget *w, GtkAllocation *alloc, struct session *sess)
{
        gnome_canvas_set_scroll_region (GNOME_CANVAS (w), 0.0, 0.0,
                                        alloc->width, alloc->height);

        gnome_canvas_item_set (sess->gui->side_box,
                               "x1", 0.0,
                               "y1", 0.0,
                               "x2", (gfloat)alloc->width,
                               "y2", (gfloat)alloc->height,
                               "width_units", 1.0, NULL);

        gnome_canvas_item_set (sess->gui->side_image,
                               "x", 0.0,
                               "y", (gfloat)(alloc->height - 192),
                               "width", 30.0,
                               "height", 192.0, NULL);
}

GtkWidget *
create_xtext_widget ()
{
	GtkWidget *xtext;

	xtext = gtk_xtext_new (prefs.dialog_indent_pixels *
			       prefs.dialog_indent_nicks,
			       TRUE);
	GTK_XTEXT (xtext)->wordwrap = prefs.dialog_wordwrap;
	GTK_XTEXT (xtext)->max_auto_indent = prefs.max_auto_indent;
	GTK_XTEXT (xtext)->auto_indent = prefs.auto_indent;
	GTK_XTEXT (xtext)->thinline = prefs.thin_separator;
	GTK_XTEXT (xtext)->max_lines = prefs.max_lines;
	GTK_XTEXT (xtext)->error_function = gtkutil_simpledialog;
	GTK_XTEXT (xtext)->urlcheck_function = maingui_word_check;

	GTK_XTEXT (xtext)->tint_red = prefs.tint_red;
	GTK_XTEXT (xtext)->tint_green = prefs.tint_green;
	GTK_XTEXT (xtext)->tint_blue = prefs.tint_blue;

	if (prefs.timestamp && prefs.dialog_indent_nicks)
		GTK_XTEXT (xtext)->time_stamp = TRUE;

	gtk_xtext_set_palette (GTK_XTEXT (xtext), colors);
	gtk_xtext_set_font (GTK_XTEXT (xtext), NULL, prefs.font_normal);
	gtk_xtext_set_background (GTK_XTEXT (xtext), NULL, 0, 0);
	
        gtk_widget_set_usize (xtext,
			      prefs.mainwindow_width - 138,
			      prefs.mainwindow_height - 106);

	return xtext;
}

gint 
gtk_kill_session_callback (GtkWidget * win, struct session * sess)
{
	kill_session_callback (sess);
	return TRUE;
}

void 
focus_in (GtkWindow * win, GtkWidget * wid, struct session *sess)
{
	if (sess && sess->gui->notebook && gtk_notebook_get_current_page (GTK_NOTEBOOK (sess->gui->notebook)) == PAGE_CHAT) {
		if (!prefs.use_server_tab)
			sess->server->front_session = sess;
		if (!sess->is_shell)
			gtk_widget_grab_focus (sess->gui->inputgad);
		else
			gtk_widget_grab_focus (sess->gui->textgad);
		menu_sess = sess;
	}
}

static void
dialog_dnd_drop (GtkWidget * widget, 
		 GdkDragContext * context, 
		 gint x,
		 gint y, 
		 GtkSelectionData * selection_data, 
		 guint info,
		 guint32 time, 
		 struct session *sess)
{
	GList *list;
	char *file, tbuf[400];

	if (!sess->channel[0])
		return;

	list = gnome_uri_list_extract_filenames (selection_data->data);
	while (list) {
		file = (char *) (list->data);
		dcc_send (sess, tbuf, sess->channel, file);
		list = list->next;
	}
	gnome_uri_list_free_strings (list);
}

GtkTargetEntry dnd_targets[] = {
	{"text/uri-list", 0, 1}
};

static char *
find_selected_nick (struct session *sess)
{
	int row;
	struct User *user;

	row = gtkutil_clist_selection (sess->gui->namelistgad);
	if (row == -1) return 0;

	user = gtk_clist_get_row_data (GTK_CLIST (sess->gui->namelistgad), row);
	if (!user) return 0;

	return user->nick;
}

static void
ul_button_rel (GtkWidget * widget, GdkEventButton * even, struct session *sess)
{
	char *nick;
	int row, col;
	char buf[67];

	if (!menu_sess)
		menu_sess = sess;

	if (!even)
		return;
	if (even->button == 3) {
		if (gtk_clist_get_selection_info (GTK_CLIST (widget), 
						  even->x, 
						  even->y, 
						  &row, 
						  &col) < 0)
			return;
		gtk_clist_unselect_all (GTK_CLIST (widget));
		gtk_clist_select_row (GTK_CLIST (widget), row, 0);

		nick = find_selected_nick (sess);
		if (nick) menu_nickmenu (sess, even, nick);

		return;
	}
	if (even->button == 2) {
		if (gtk_clist_get_selection_info (GTK_CLIST (widget), 
						  even->x, 
						  even->y, 
						  &row, 
						  &col) != -1) {
			gtk_clist_select_row (GTK_CLIST (widget), row, col);
			nick = find_selected_nick (sess);
			if (nick) {
				snprintf (buf, sizeof (buf), "%s: ", nick);
				gtk_entry_set_text (
					GTK_ENTRY (sess->gui->inputgad), 
					buf);
				gtk_widget_grab_focus (sess->gui->inputgad);
			}
		}
	}
}

static void
maingui_userlist_selected (GtkWidget * clist, gint row, gint column,
			   GdkEventButton * even)
{
	char *nick;
	void *unused;

	if (even) {
		if (even->type == GDK_2BUTTON_PRESS) {
			if (prefs.doubleclickuser[0]) {
				
				if (gtk_clist_get_cell_type
				    (GTK_CLIST (clist), row, column)
				    == GTK_CELL_PIXTEXT) {
					gtk_clist_get_pixtext (GTK_CLIST
							       (clist), row,
							       column, &nick,
							       (guint8 *) &
							       unused,
							       (GdkPixmap **) &
							       unused,
							       (GdkBitmap **) &
							       unused);
				} else
					gtk_clist_get_text (GTK_CLIST (clist),
							    row, column, &nick);

				nick_command_parse (menu_sess,
						    prefs.doubleclickuser, nick,
						    nick);
			}
		} else {
			if (!(even->state & GDK_SHIFT_MASK)) {
				gtk_clist_unselect_all (GTK_CLIST (clist));
				gtk_clist_select_row (GTK_CLIST (clist), row,
						      column);
			}
		}
	}
}

#define BANNER_FONT \
	"font", \
	"-adobe-helvetica-bold-r-normal-*-18-*-*-*-p-*-iso8859-1"
#define BANNER_FONTSET \
	"fontset", \
	"-adobe-helvetica-bold-r-normal-*-18-*-*-*-p-*-iso8859-1,*-r-*"

void 
init_canvas (struct session *sess)
{
        GdkPixbuf *pb;
        GnomeCanvasItem *root;
        GnomeCanvas *canvas;

        canvas = GNOME_CANVAS (sess->gui->title_canvas);
        root = GNOME_CANVAS_ITEM (gnome_canvas_root (GNOME_CANVAS (canvas)));

        sess->gui->title_box = 
                gnome_canvas_item_new (GNOME_CANVAS_GROUP (root),
                                       gnome_canvas_rect_get_type (),
                                       "fill_color", "#314a82",
                                       "outline_color", "#314a82",
				       NULL);

        sess->gui->banner =
                gnome_canvas_item_new (GNOME_CANVAS_GROUP (root),
                                       gnome_canvas_text_get_type (),
                                       "fill_color", "#d6d6d6",
				       BANNER_FONT,
				       BANNER_FONTSET,
                                       "anchor", GTK_ANCHOR_WEST,
				       "text", "MonkeyTalk",
                                       NULL);

	pb = gdk_pixbuf_new_from_file (DATADIR "/monkeytalk/talk-chanlist.png");

        sess->gui->chanlist_btn = 
		gnome_canvas_item_new (GNOME_CANVAS_GROUP (root),
				       GNOME_TYPE_CANVAS_PIXBUF,
				       "pixbuf", pb, 
				       NULL);

	pb = gdk_pixbuf_new_from_file (DATADIR "/monkeytalk/talk-userlist.png");

        sess->gui->userlist_btn = 
		gnome_canvas_item_new (GNOME_CANVAS_GROUP (root),
				       GNOME_TYPE_CANVAS_PIXBUF,
				       "pixbuf", pb, 
				       NULL);

	gtk_signal_connect (GTK_OBJECT (sess->gui->chanlist_btn), 
			    "event",
			    GTK_SIGNAL_FUNC (click_chanlist), 
			    sess);

	gtk_signal_connect (GTK_OBJECT (sess->gui->userlist_btn), 
			    "event",
			    GTK_SIGNAL_FUNC (click_userlist), 
			    sess);

	sess->gui->tooltip = gtk_tooltips_new ();
	gtk_tooltips_disable (sess->gui->tooltip);

        /*******************************************************/
        canvas = GNOME_CANVAS (sess->gui->side_canvas);
        root = GNOME_CANVAS_ITEM (gnome_canvas_root (GNOME_CANVAS (canvas)));

        sess->gui->side_box = 
		gnome_canvas_item_new (GNOME_CANVAS_GROUP (root),
				       gnome_canvas_rect_get_type (),
				       "fill_color", "#314a82",
				       "outline_color", "#314a82", 
				       NULL);

        pb = gdk_pixbuf_new_from_file (DATADIR "/monkeytalk/talk-blocks.png");

        sess->gui->side_image =
                gnome_canvas_item_new (GNOME_CANVAS_GROUP (root),
                                       GNOME_TYPE_CANVAS_PIXBUF,
                                       "pixbuf", pb, 
				       NULL);
}

static void
goto_next (struct session *sess)
{	
	int page = gtk_notebook_get_current_page (GTK_NOTEBOOK (sess->gui->notebook)) + 1;
	
	if (page == LAST_PAGE) {
		page = 0;
	}

	gtk_notebook_set_page (GTK_NOTEBOOK (sess->gui->notebook),
			       page);
}

static gboolean
check_nick (struct session *sess, char *nick) 
{
	GtkWidget *dlg = NULL;
	char *p;

	if (!g_strcasecmp (nick, "root")) {
		dlg = gnome_error_dialog_parented (_("Sorry, you can not use \"root\" as your nickname."), 
						   GTK_WINDOW (sess->gui->window));
	} 

	if (!dlg) {
		for (p = nick; *p != 0; p++) {
			if (isspace (*p)) {
				dlg = gnome_error_dialog_parented (_("Sorry, you cannot use spaces in your nickname."),
								   GTK_WINDOW (sess->gui->window));	
				break;
			}
		}
	}
	
	if (dlg) {
		gnome_dialog_run_and_close (GNOME_DIALOG (dlg));
		return FALSE;
	} else {
		return TRUE;
	}
}

static char *
strip_whitespace (const char *orig)
{
	char *ret;
	char *cpy = g_strdup (orig);
	char *start = NULL;
	char *p;
	
	for (p = cpy; *p != 0; p++) {
		if (!isspace (*p) && !start) {
			start = p;
		}	
	}

	p--;
	for (; isspace (*p); p--) {
	}
	
	*(++p) = 0;
	
	ret = g_strdup (start ? start : "");
	g_free (cpy);
	return ret;
}

static void
nick_next_cb (GtkWidget *widget, gpointer data)
{
	struct session *sess = data;
	char *nick = gtk_entry_get_text (GTK_ENTRY (sess->gui->nick_entry));
	
	if (nick) {
		nick = strip_whitespace (nick);

		if (check_nick (sess, nick)) {
			gnome_config_set_string ("/monkeytalk/Default/nick", nick);

			
			strncpy (prefs.nick1, nick, 63);
			strncpy (prefs.nick2, nick, 62);
			strcat (prefs.nick2, "_");
			strncpy (prefs.nick3, nick, 61);
			strcat (prefs.nick3, "__");
			strncpy (prefs.realname, nick, 63);
			strncpy (prefs.username, nick, 63);

			fe_set_nick (sess->server, prefs.nick1);

			gnome_config_sync ();
			
			goto_next (sess);
			connect_to_support_channel (sess);
		}
		g_free (nick);
	}

}

static void
next_cb (GtkWidget *widget, gpointer data)
{
	struct session *sess = data;
	
	switch (gtk_notebook_get_current_page (GTK_NOTEBOOK (sess->gui->notebook))) {
	case PAGE_NOTICE :
		goto_next (sess);
		break;
	case PAGE_NICK :
		nick_next_cb (widget, data);
		break;
	case PAGE_CHAT :
		break;
	}
}

static void
prev_cb (GtkWidget *widget, gpointer data)
{
	struct session *sess = data;
	
	gtk_notebook_set_page (GTK_NOTEBOOK (sess->gui->notebook),
			       gtk_notebook_get_current_page (GTK_NOTEBOOK (sess->gui->notebook)) - 1);
}

static void
notice_data_cb (GnomeTransfer * tran,
		char *buf, 
		int size,
		GtkWidget *notice)
{
	GtkHTMLStream *stream = gtk_object_get_data (GTK_OBJECT (notice),
						     "stream");
	
	gtk_html_write (GTK_HTML (notice), stream, buf, size);
	
}

static void
notice_done_cb (GnomeTransfer * tran,
		GnomeTransferStatus status, 
		GtkWidget *notice)
{
	GtkHTMLStream *stream = gtk_object_get_data (GTK_OBJECT (notice),
						     "stream");
	
	gtk_html_end (GTK_HTML (notice), stream, 
		      status == GNOME_TRANSFER_COMPLETE 
		      ? GTK_HTML_STREAM_OK
		      : GTK_HTML_STREAM_ERROR);
}

static void
notice_link_clicked_cb (GtkHTML *html, const gchar *url,
			struct session *sess)
{
	if (!strcmp (url, "[NEXT]")) {
		goto_next (sess);
	} else {
		gnome_url_show (url);
	}
}

#define proxy_key "/system/gnome-vfs/use-http-proxy"
#define pauth_key "/system/gnome-vfs/use-http-proxy-authorization"
#define phost_key "/system/gnome-vfs/http-proxy-host"
#define pport_key "/system/gnome-vfs/http-proxy-port"
#define puser_key "/system/gnome-vfs/http-proxy-authorization-user"
#define ppass_key "/system/gnome-vfs/http-proxy-authorization-password"

static GtkWidget *
create_notice_page (struct session *sess)
{
	GtkWidget *notice;
	GtkWidget *scrolled;
	GtkWidget *frame;
	gchar *url, *proxyurl = NULL;
	gchar *proxyhost = NULL;
	gushort proxyport = 8080;
	gchar *proxyuser = NULL, *proxypass = NULL;
	GnomeTransfer *tran = gnome_transfer_new ();
	GtkHTMLStream *stream;
	GConfClient *client;

	notice = gtk_html_new ();

	gtk_signal_connect (GTK_OBJECT (notice), "link_clicked", 
			    GTK_SIGNAL_FUNC (notice_link_clicked_cb), sess);
	
	stream = gtk_html_begin (GTK_HTML (notice));
	gtk_object_set_data (GTK_OBJECT (notice), "stream", stream);

	gtk_signal_connect (GTK_OBJECT (tran),
			    "file_data",
			    GTK_SIGNAL_FUNC (notice_data_cb), 
			    notice);
	gtk_signal_connect (GTK_OBJECT (tran), 
			    "file_done",
			    GTK_SIGNAL_FUNC (notice_done_cb),
			    notice);
	
	url = construct_url ("http://monkeytalk.ximian.com/notice.html");

	client = gconf_client_get_default ();

	if (gconf_client_get_bool (client, proxy_key, NULL)) {
		proxyhost = gconf_client_get_string (client, phost_key, NULL);
		proxyport = gconf_client_get_int (client, pport_key, NULL);
		proxyurl = g_strdup_printf ("http://%s:%d", proxyhost,
					    proxyport);
		gnome_transfer_proxy (tran, proxyurl);
		if (gconf_client_get_bool (client, pauth_key, NULL)) {
			proxyuser = gconf_client_get_string (client, puser_key,
							     NULL);
			proxypass = gconf_client_get_string (client, ppass_key,
							     NULL);
			gnome_transfer_proxy_password (tran, proxyuser,
						       proxypass);
		}
	}

	if (gnome_transfer_begin (tran, url) == -1) {
		char *txt;
		GtkWidget *dlg;

		txt = g_strdup_printf (_("Connection failed: %s"), 
				       gnome_transfer_reason (tran));
		dlg = gnome_error_dialog (txt);
		g_free (txt);
		gnome_dialog_run_and_close (GNOME_DIALOG (dlg));
		exit (1);
	}

	g_free (proxyhost);
	g_free (proxyurl);
	g_free (proxyuser);
	g_free (proxypass);
	g_free (url);

	gtk_widget_show (notice);

	scrolled = gtk_scrolled_window_new (NULL, NULL);
	gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (scrolled),
					GTK_POLICY_AUTOMATIC, 
					GTK_POLICY_AUTOMATIC);
	gtk_widget_show (scrolled);
	gtk_container_add (GTK_CONTAINER (scrolled), notice);

	frame = gtk_frame_new (NULL);
	gtk_frame_set_shadow_type (GTK_FRAME (frame), GTK_SHADOW_IN);
	gtk_container_add (GTK_CONTAINER (frame), scrolled);

	gtk_container_set_border_width (GTK_CONTAINER (frame),
					GNOME_PAD_SMALL);

	return frame;
}

static void
nick_entry_changed_cb (GtkWidget *entry, gpointer data)
{
	struct session *sess = data;
	GtkWidget *btn;
	char *text;
	char *p;
	gboolean enabled = FALSE;
	
	btn = sess->gui->next_btn;

	if (btn && gtk_notebook_get_current_page (GTK_NOTEBOOK (sess->gui->notebook)) == PAGE_NICK)  {
		text = gtk_entry_get_text (GTK_ENTRY (entry));
		
		if (text) {
			for (p = text; *p != 0; p++) {
				if (!isspace (*p)) {
					enabled = TRUE;
				}
			}
		}
		
		gtk_widget_set_sensitive (btn, enabled);
	}
}


static GtkWidget *
create_nick_page (struct session *sess) 
{
	GtkWidget *vbox;
	GtkWidget *explanation;
	GtkWidget *hbox;
	GtkWidget *nick_label;
	GtkWidget *entry;
	char *nick_val;
	
	vbox = gtk_vbox_new (FALSE, GNOME_PAD);
	gtk_container_set_border_width (GTK_CONTAINER (vbox), 
					GNOME_PAD_SMALL);
	explanation = gtk_label_new (_("Your MonkeyTalk nick identifies you to other users of the system. \nPlease type the nickname you would like to use here."));
	
	gtk_widget_show (explanation);
	gtk_box_pack_start (GTK_BOX (vbox), explanation, FALSE, FALSE, 0);
	
	hbox = gtk_hbox_new (FALSE, GNOME_PAD_SMALL);
	gtk_widget_show (hbox);
	gtk_box_pack_start (GTK_BOX (vbox), hbox, FALSE, FALSE, 0);
	
	nick_label = gtk_label_new (_("Nick"));
	gtk_widget_show (nick_label);
	gtk_box_pack_start (GTK_BOX (hbox), nick_label, FALSE, FALSE, 0);
	
	entry = gtk_entry_new_with_max_length (60);
	gtk_widget_show (entry);
	gtk_box_pack_start (GTK_BOX (hbox), entry, FALSE, FALSE, 0);
	sess->gui->nick_entry = entry;	

	gtk_signal_connect (GTK_OBJECT (entry), "changed",
			    GTK_SIGNAL_FUNC (nick_entry_changed_cb), sess);

	nick_val = prefs.nick1;
	if (nick_val) {
		gtk_entry_set_text (GTK_ENTRY (entry), nick_val);
	}

	gtk_container_set_border_width (GTK_CONTAINER (vbox),
					GNOME_PAD_SMALL);

	return vbox;
}

static GtkWidget *
create_chat_page (struct session *sess)
{
	GtkWidget *vbox23;
	GtkWidget *notebook_holder;
	GtkWidget *vbox27;
	GtkWidget *vbox_foo;
	GtkWidget *hbox_foo;
	GtkWidget *hbox49;
	GtkWidget *entry_frame;
	GtkWidget *pixmap;
	GtkWidget *nickname;
	GtkWidget *entry;
	GtkWidget *send_button;
	GtkWidget *hbox50;
	GtkWidget *xtext_frame;
	GtkWidget *xtext;
	GtkWidget *vscrollbar;
	GtkWidget *user_vbox = NULL;
	GtkWidget *user_clist = NULL;

	vbox23 = gtk_vbox_new (FALSE, 0);
	gtk_widget_show (vbox23);

	if (!sess->is_dialog) {
		GtkWidget *canvas1;
		GtkWidget *hbox37;
		GtkWidget *vbox24;
		GtkWidget *canvas2;

		gdk_rgb_init ();
		gtk_widget_push_visual (gdk_rgb_get_visual ());
		gtk_widget_push_colormap (gdk_rgb_get_cmap ());
		canvas1 = gnome_canvas_new_aa ();
		gtk_widget_pop_colormap ();
		gtk_widget_pop_visual ();
		gtk_widget_show (canvas1);
		gtk_box_pack_start (GTK_BOX (vbox23), canvas1, FALSE, FALSE, 0);
		gtk_widget_set_usize (canvas1, 450, 45);
		gnome_canvas_set_scroll_region (GNOME_CANVAS (canvas1), 
						0, 
						0, 
						100, 
						100);

		hbox37 = gtk_hbox_new (FALSE, 0);
		gtk_widget_show (hbox37);
		gtk_box_pack_start (GTK_BOX (vbox23), hbox37, TRUE, TRUE, 0);

		vbox24 = gtk_vbox_new (FALSE, 0);
		gtk_widget_show (vbox24);
		gtk_box_pack_start (GTK_BOX (hbox37), vbox24, FALSE, FALSE, 0);

		gdk_rgb_init ();
		gtk_widget_push_visual (gdk_rgb_get_visual ());
		gtk_widget_push_colormap (gdk_rgb_get_cmap ());
		canvas2 = gnome_canvas_new_aa ();
		gtk_widget_pop_colormap ();
		gtk_widget_pop_visual ();
		gtk_widget_show (canvas2);
		gtk_box_pack_start (GTK_BOX (vbox24), canvas2, TRUE, TRUE, 0);
		gtk_widget_set_usize (canvas2, 24, 250);
		gnome_canvas_set_scroll_region (GNOME_CANVAS (canvas2), 
						0, 
						0, 
						100, 
						100);

		gtk_signal_connect_after (
			GTK_OBJECT (canvas1), 
			"size_allocate",
			GTK_SIGNAL_FUNC (title_configure_size), 
			sess);

		gtk_signal_connect_after (
			GTK_OBJECT (canvas2), 
			"size_allocate",
			GTK_SIGNAL_FUNC (side_configure_size), 
			sess);

		GTK_WIDGET_UNSET_FLAGS (canvas1, GTK_CAN_FOCUS);
		GTK_WIDGET_UNSET_FLAGS (canvas2, GTK_CAN_FOCUS);

		sess->gui->title_canvas = GNOME_CANVAS (canvas1);
		sess->gui->side_canvas = GNOME_CANVAS (canvas2);

		init_canvas (sess);

		notebook_holder = hbox37;
	} else {
		notebook_holder = vbox23;
	}

	vbox27 = gtk_vbox_new (FALSE, 1);
	gtk_widget_show (vbox27);
	gtk_container_add (GTK_CONTAINER (notebook_holder), vbox27);
	gtk_container_set_border_width (GTK_CONTAINER (vbox27), 2);

	hbox50 = gtk_hbox_new (FALSE, 1);
	gtk_widget_show (hbox50);
	gtk_box_pack_start (GTK_BOX (vbox27), hbox50, TRUE, TRUE, 0);

	xtext_frame = gtk_frame_new (NULL);
	gtk_frame_set_shadow_type (GTK_FRAME (xtext_frame), 
				   GTK_SHADOW_ETCHED_IN);
	gtk_widget_show (xtext_frame);
	gtk_box_pack_start (GTK_BOX (hbox50), xtext_frame, TRUE, TRUE, 0);


	xtext = create_xtext_widget ();
	gtk_widget_show (xtext);
	gtk_container_add (GTK_CONTAINER (xtext_frame), xtext);
	GTK_WIDGET_UNSET_FLAGS (xtext, GTK_CAN_FOCUS);
	GTK_WIDGET_UNSET_FLAGS (xtext, GTK_CAN_DEFAULT);

	gtk_object_set_user_data (GTK_OBJECT (xtext), sess);

	gtk_drag_dest_set (xtext, 
			   GTK_DEST_DEFAULT_ALL,
			   dnd_targets, 
			   1,
			   GDK_ACTION_MOVE | GDK_ACTION_COPY | GDK_ACTION_LINK);

	gtk_signal_connect (GTK_OBJECT (xtext), 
			    "word_click",
			    GTK_SIGNAL_FUNC (maingui_word_clicked), 
			    sess);

	if (sess->is_dialog)
		gtk_signal_connect (GTK_OBJECT (xtext),
				    "drag_data_received",
				    GTK_SIGNAL_FUNC (dialog_dnd_drop), 
				    sess);

	vscrollbar = gtk_vscrollbar_new (GTK_XTEXT (xtext)->adj);
	gtk_box_pack_start (GTK_BOX (hbox50), vscrollbar, FALSE, FALSE, 0);
	GTK_WIDGET_UNSET_FLAGS (vscrollbar, GTK_CAN_FOCUS);
	gtk_widget_show (vscrollbar);

	if (!sess->is_dialog) {
		user_vbox = gtk_vbox_new (FALSE, 0);
		gtk_box_pack_end (GTK_BOX (hbox50), 
				  user_vbox, 
				  FALSE, 
				  FALSE, 
				  0);

		user_clist = 
			gtkutil_clist_new (1, 
					   0, 
					   user_vbox, 
					   GTK_POLICY_AUTOMATIC,
					   maingui_userlist_selected, 
					   sess, 
					   0, 
					   0,
					   GTK_SELECTION_MULTIPLE);
		GTK_WIDGET_UNSET_FLAGS (user_clist, GTK_CAN_FOCUS);

		gtk_clist_set_shadow_type (GTK_CLIST (user_clist), 
					   GTK_SHADOW_ETCHED_IN);

		/* set up drops */
		gtk_drag_dest_set (
			 user_clist, 
			 GTK_DEST_DEFAULT_ALL,
			 dnd_targets, 
			 1,
			 GDK_ACTION_MOVE | GDK_ACTION_COPY | GDK_ACTION_LINK);

		gtk_signal_connect (GTK_OBJECT (user_clist), 
				    "drag_motion",
				    GTK_SIGNAL_FUNC (userlist_dnd_motion), 
				    0);
		gtk_signal_connect (GTK_OBJECT (user_clist), 
				    "drag_leave",
				    GTK_SIGNAL_FUNC (userlist_dnd_leave), 
				    0);
		gtk_signal_connect (GTK_OBJECT (user_clist),
				    "drag_data_received",
				    GTK_SIGNAL_FUNC (userlist_dnd_drop), 
				    sess);

		gtk_signal_connect (GTK_OBJECT (user_clist),
				    "button_press_event",
				    GTK_SIGNAL_FUNC (ul_button_rel), 
				    sess);

		gtk_clist_set_selection_mode (GTK_CLIST (user_clist),
					      GTK_SELECTION_MULTIPLE);
		gtk_clist_set_column_width (GTK_CLIST (user_clist), 0, 10);
		gtk_widget_set_usize (user_clist, 115, 0);

		if (!prefs.hideuserlist) {
			gtk_widget_show_all (user_vbox);
		}
		sess->userlisthidden = prefs.hideuserlist;
	}

	entry_frame = gtk_frame_new (NULL);
	gtk_frame_set_shadow_type (GTK_FRAME (entry_frame), 
				   GTK_SHADOW_ETCHED_IN);
	gtk_widget_show (entry_frame);
	gtk_box_pack_start (GTK_BOX (vbox27), entry_frame, FALSE, FALSE, 0);

	vbox_foo = gtk_vbox_new (FALSE, 0);
	gtk_widget_show (vbox_foo);
	gtk_container_add (GTK_CONTAINER (entry_frame), vbox_foo);
	
	hbox49 = gtk_hbox_new (FALSE, 0);
	gtk_widget_show (hbox49);
	gtk_box_pack_start (GTK_BOX (vbox_foo), hbox49, FALSE, FALSE, 8);

	hbox_foo = gtk_hbox_new (FALSE, 4);
	gtk_widget_show (hbox_foo);
	gtk_box_pack_start (GTK_BOX (hbox49), hbox_foo, FALSE, FALSE, 8);


	pixmap = gtk_pixmap_new (pix_voice, mask_voice);
	gtk_widget_show (pixmap);
	gtk_box_pack_start (GTK_BOX (hbox_foo), pixmap, FALSE, FALSE, 0);


	nickname = gtk_label_new (sess->server->nick);
	gtk_widget_show (nickname);
	gtk_box_pack_start (GTK_BOX (hbox_foo), nickname, FALSE, FALSE, 0);
	gtk_misc_set_alignment (GTK_MISC (nickname), 0.5, 0.3);


	entry = gtk_entry_new_with_max_length (2048);
	gtk_widget_show (entry);
	gtk_box_pack_start (GTK_BOX (hbox49), entry, TRUE, TRUE, 0);
	gtk_widget_grab_focus (entry);

	gtk_signal_connect (GTK_OBJECT (entry), 
			    "key_press_event",
			    GTK_SIGNAL_FUNC (key_handle_key_press), 
			    sess);

	send_button = gtk_button_new_with_label (_("Send"));
	gtk_widget_show (send_button);
	gtk_box_pack_start (GTK_BOX (hbox49), send_button, FALSE, FALSE, 8);
	GTK_WIDGET_UNSET_FLAGS (send_button, GTK_CAN_FOCUS);

	gtk_signal_connect (GTK_OBJECT (send_button), 
			    "clicked",
			    GTK_SIGNAL_FUNC (key_handle_send), 
			    sess);

	sess->gui->type = 
		sess->is_dialog ? TALK_WINDOW_CHANNEL : TALK_WINDOW_PRIVATE;
	sess->gui->textgad = xtext;
	sess->gui->inputgad = entry;
	sess->gui->vscrollbar = vscrollbar;
	sess->gui->bar = NULL;
	sess->gui->op_box = hbox49;
	sess->gui->op_xpm = pixmap;
	sess->gui->namelistgad = user_clist;
	sess->gui->userlistbox = user_vbox;
	sess->gui->send_button = send_button;
	sess->gui->nickname = nickname;

	return vbox23;
}

static GtkWidget *
create_nav_buttons (struct session *sess)
{
	GtkWidget *button_box;
	GtkWidget *btn;
	
	button_box = gtk_hbutton_box_new ();

	btn = gnome_stock_button (GNOME_STOCK_BUTTON_PREV);
	gtk_widget_show (btn);	
	gtk_container_add (GTK_CONTAINER (button_box), btn);
	gtk_signal_connect (GTK_OBJECT (btn), "clicked",
			    GTK_SIGNAL_FUNC (prev_cb), sess);
	sess->gui->prev_btn = btn;
	gtk_widget_set_sensitive (btn, FALSE);

	btn = gnome_stock_button (GNOME_STOCK_BUTTON_NEXT);
	gtk_widget_show (btn);	
	gtk_container_add (GTK_CONTAINER (button_box), btn);
	gtk_signal_connect (GTK_OBJECT (btn), "clicked",
			    GTK_SIGNAL_FUNC (next_cb), sess);
	sess->gui->next_btn = btn;
	
	btn = gnome_stock_button (GNOME_STOCK_BUTTON_CANCEL);
	gtk_widget_show (btn);
	gtk_signal_connect (GTK_OBJECT (btn), "clicked",
			    GTK_SIGNAL_FUNC (fe_exit), NULL);
	gtk_container_add (GTK_CONTAINER (button_box), btn);

	gtk_button_box_set_layout (GTK_BUTTON_BOX (button_box),
				   GTK_BUTTONBOX_END);
	gtk_container_set_border_width (GTK_CONTAINER (button_box),
					GNOME_PAD_SMALL);

	return button_box;
}

void
notebook_switch_page_cb (GtkWidget *w, GtkNotebookPage *page,
			 gint page_num, gpointer data)
{
	struct session *sess = data;
	
	switch (page_num) {
	case PAGE_NOTICE: 
		if (sess->gui->prev_btn) {
			gtk_widget_set_sensitive (sess->gui->prev_btn, FALSE);
 			gtk_widget_show (sess->gui->page_nav_box);
		}
		
		break;
	case PAGE_NICK :
		if (sess->gui->next_btn) {
			gtk_widget_set_sensitive (sess->gui->prev_btn, TRUE);
		}
		if (sess->gui->nick_entry) {
			gtk_widget_grab_focus (sess->gui->nick_entry);
			gtk_widget_show (sess->gui->page_nav_box);
		}
		break;
	case PAGE_CHAT :
		if (sess->gui->page_nav_box) {
			gtk_widget_hide (sess->gui->page_nav_box);
		}
		break;
	default :
		break;
	}
}
 
static void
create_config (struct session *sess) 
{
	GtkWidget *notice_page;
	GtkWidget *nick_page;
	GtkWidget *bbox;

	bbox = create_nav_buttons (sess);
	sess->gui->page_nav_box = bbox;
	gtk_box_pack_end (GTK_BOX (sess->gui->main_vbox), 
			  bbox, FALSE, FALSE, 0);
	
	notice_page = create_notice_page (sess);
	sess->gui->pages[PAGE_NOTICE] = notice_page;
	gtk_notebook_append_page (GTK_NOTEBOOK (sess->gui->notebook),
				  notice_page, NULL);

	nick_page = create_nick_page (sess);
	sess->gui->pages[PAGE_NICK] = nick_page;
	gtk_notebook_append_page (GTK_NOTEBOOK (sess->gui->notebook),
				  nick_page, NULL);
}

static void 
show_config (struct session *sess)
{
	int i;

	if (!sess->gui->page_nav_box) {
		create_config (sess);
	}
	
	for (i = 1; i < LAST_PAGE; i++) {
		gtk_widget_show (sess->gui->pages[i]);
	}
	gtk_widget_show (sess->gui->page_nav_box);
	gtk_notebook_set_page (GTK_NOTEBOOK (sess->gui->notebook),
			       PAGE_NOTICE);
}

void
create_talk_window_new (struct session *sess)
{
	GtkWidget *talk_window;
	GtkWidget *chat_page;
	GtkWidget *notebook;
	GtkWidget *vbox;

	static gboolean first_run = 1;
	
	sess->is_tab = FALSE;
	sess->logfd = -1;
	sess->new_data = FALSE;
	sess->gui = g_new0 (TalkWindow, 1);

	talk_window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
	sess->gui->window = talk_window;
	gtk_widget_set_usize (talk_window, 
			      sess->is_dialog ? 440 : 540, 
			      sess->is_dialog ? 256 : 400);
	gtk_window_set_title (GTK_WINDOW (talk_window), _("MonkeyTalk"));
	gtk_window_set_policy (GTK_WINDOW (talk_window), 
			       TRUE,
			       TRUE, 
			       FALSE);
	gnome_window_icon_set_default_from_file (ICONDIR "/monkeytalk.png");
	gnome_window_icon_set_from_default (GTK_WINDOW (talk_window));

	palette_alloc (talk_window);

	gtk_signal_connect (GTK_OBJECT (talk_window), 
			    "destroy",
			    GTK_SIGNAL_FUNC (gtk_kill_session_callback),
			    sess);

	gtk_signal_connect (GTK_OBJECT (talk_window),
			    "focus_in_event",
			    GTK_SIGNAL_FUNC (focus_in),
			    sess);	

	notebook = gtk_notebook_new ();
	sess->gui->notebook = notebook;
	gtk_notebook_set_show_tabs (GTK_NOTEBOOK (notebook), FALSE);
	gtk_notebook_set_show_border (GTK_NOTEBOOK (notebook), FALSE);
	gtk_signal_connect (GTK_OBJECT (notebook), "switch_page",
			    GTK_SIGNAL_FUNC (notebook_switch_page_cb), sess);

	chat_page = create_chat_page (sess);
	sess->gui->pages[PAGE_CHAT] = chat_page;
	gtk_notebook_append_page (GTK_NOTEBOOK (notebook), chat_page, NULL);

	vbox = gtk_vbox_new (FALSE, GNOME_PAD_SMALL);
	sess->gui->main_vbox = vbox;
	gtk_widget_show (vbox);
	
	gtk_box_pack_start (GTK_BOX (vbox), notebook, TRUE, TRUE, 0);

	gtk_container_add (GTK_CONTAINER (talk_window), vbox);
	gtk_widget_show (notebook);

	if (first_run) {
		first_run = FALSE;
		show_config (sess);
	}

	gtk_widget_show (talk_window);	

	fe_set_title (sess);
}

int
fe_args (int argc, char **argv) 
{
	GError *gconf_error;

#ifdef ENABLE_NLS
	bindtextdomain (PACKAGE, PREFIX "/share/locale");
	textdomain (PACKAGE);
#endif

	gtk_set_locale ();

	gnome_init ("MonkeyTalk", VERSION, argc, argv);

	if (argc > 1) {
		support_keyword = argv [1];
	}

	if (!gconf_init (argc, argv, &gconf_error)) {
		g_assert (gconf_error != NULL);
		g_error ("GConf init failed:\n  %s", gconf_error->message);
		return FALSE;
	}

	return 1;
}

void
fe_init (void)
{
	prefs.use_server_tab = 0;
	prefs.topicbar = 0;
	prefs.lagometer = 0;
	prefs.userlistbuttons = 0;
	prefs.timestamp = 0;
	prefs.nickgad = 1;
	prefs.hideuserlist = gnome_config_get_bool ("/monkeytalk/Default/hideuserlist=true");
	prefs.tabchannels = 0;
	prefs.hidemenu = 1;
	prefs.showhostname_in_userlist = 0;
	prefs.dialog_indent_nicks = 1;
	prefs.indent_nicks = 1;
	prefs.skipmotd = 1;

	prefs.mail_check = 0;
	prefs.nickcompletion = 0;
	prefs.show_separator = 0;
	prefs.skipserverlist = 1;
	prefs.timestamp_logs = 1;
	prefs.treeview = 0;
	prefs.wallops = 0;
	prefs.wordwrap = 0;
	prefs.autodccchat = 1;
	prefs.autoreconnect = 1;
	prefs.autodccsend = 0;
	prefs.colorednicks = 1;

	/* proxy settings?? */

	pixmaps_init ();
}

void fe_main (void)
{
	gtk_main ();
}

void fe_cleanup (void)
{
}

void 
fe_exit (void)
{
	gtk_main_quit ();
}

int 
fe_timeout_add (int interval, void *callback, void *userdata)
{
	return gtk_timeout_add (interval, (GtkFunction) callback, userdata);
}

void 
fe_timeout_remove (int tag)
{
	gtk_timeout_remove (tag);
}

void 
fe_set_title (struct session *sess)
{
	char tbuf[200];

	if (!sess->server->connected && !sess->is_dialog)
		strcpy (tbuf, "MonkeyTalk: Not Connected");
	else if (sess->channel[0] == 0 || sess->is_server)
		strcpy (tbuf, "MonkeyTalk: Connected");
	else
		strcpy (tbuf, "MonkeyTalk");

	gtk_window_set_title (GTK_WINDOW (sess->gui->window), tbuf);
}

static void
transfer_data_cb (GnomeTransfer * tran,
		  char *buf, 
		  int size,
		  GByteArray **dest)
{
	if (!*dest) *dest = g_byte_array_new ();

	g_byte_array_append (*dest, buf, size);
}

static void
default_channel_transfer_done_cb (GnomeTransfer * tran,
				  GnomeTransferStatus status, 
				  struct session *sess)
{
	gchar **lines = NULL, **iter;

	if (status == GNOME_TRANSFER_FAILED) goto ERROR;
	if (support_server_result->len == 0) goto ERROR;

	g_byte_array_append (support_server_result, "\0", 1);

	lines = g_strsplit (support_server_result->data, "\n\n\n", 0);

	if (!lines) goto ERROR;
	else if (!lines [0] || !lines [1]) {
		g_strfreev (lines);
		goto ERROR;
	}

	for (iter = lines; *iter; iter++) {
		gchar *colon, *key, *val;

		colon = strchr (*iter, ':');
		if (!colon) break;

		*colon = '\0';

		key = g_strstrip (*iter);
		val = g_strstrip (colon + 1);

		if (!g_strcasecmp (key, "server")) {
			char *colon2 = strchr (val, ':');
			if (colon2) {
				support_server_port = atoi (colon2 + 1);
				*colon2 = '\0';
			}

			support_server = g_strdup (val);
		} else if (!g_strcasecmp (key, "channel"))
			support_channel = g_strdup (val);
		else if (!g_strcasecmp (key, "greeting"))
			support_greeting = g_strdup (val);
	}

	if (!support_server || !support_channel) goto ERROR;

	g_byte_array_free (support_server_result, TRUE);
	g_strfreev (lines);

	sess->server->front_session = sess;
	strcpy (sess->willjoinchannel, support_channel);

	connect_server (sess, 
			support_server, 
			support_server_port, 
			FALSE);

	return;

 ERROR:
	gnome_dialog_run_and_close (
		gtkutil_simpledialog ("Unable to contact Ximian Support "
				      "server.\n"));
	exit (1);
}

/* These are all the characters defined as safe in the url spec */
#define IS_SAFE(c) (isalnum(c) || c == '$' || c == '-' || c == '_' || c == '.' || c == '&' || c == '+' || c == '!' || c == '*' || c == '=' || c == '\'' || c == '(' || c == ')' || c == ',' || c == ';' || c == '/' || c == '#' || c == '?' || c == ':' || c == '@')

static gchar *
escape_url (gchar *url) 
{
	char *p, *ret;
	GString *str = g_string_new ("");

	for (p = url;*p != 0; p++) {
		if (IS_SAFE (*p)) {
			g_string_append_c (str, *p);
		} else {
			char buf[4];
			g_snprintf (buf, 4, "%%%2hhx", *p);
			g_string_append (str, buf);
		}
	}
	ret = str->str;
	g_string_free (str, FALSE);	
	return ret;
}

static gchar *
construct_url (gchar *url)
{
	gchar *tmp, *ret, *lang;

	lang = getenv ("GDM_LANG");
	if (!lang || !strcmp (lang, "(null)")) lang = getenv ("LANG");

	tmp = g_strconcat (url,
			   "?",
			   "talkversion=" VERSION,
			   lang ? "&lang=" : "",
			   lang ? lang : "",
			   support_keyword ? "&appname=" : "",
			   support_keyword ? support_keyword : "",
			   NULL);

	ret = escape_url (tmp);
	g_free (tmp);

	return ret;
}

static void 
begin_default_channel_query (struct session *sess)
{
	gchar *url, *proxyurl = NULL;
	gchar *proxyhost = NULL;
	gint proxyport;
	gchar *proxyuser = NULL, *proxypass = NULL;
	GConfClient *client;
	GnomeTransfer *tran = gnome_transfer_new ();

	gtk_signal_connect (GTK_OBJECT (tran),
			    "file_data",
			    GTK_SIGNAL_FUNC (transfer_data_cb), 
			    &support_server_result);

	gtk_signal_connect (GTK_OBJECT (tran),
			    "file_done",
			    GTK_SIGNAL_FUNC (default_channel_transfer_done_cb), 
			    sess);

	if (g_getenv ("MONKEYTALK_TEST")) {
		url = construct_url ("http://monkeytalk.ximian.com/cgi-bin/test-support");
	} else {		
		url = construct_url ("http://monkeytalk.ximian.com/cgi-bin/get-support");
	}
	
	client = gconf_client_get_default ();

	if (gconf_client_get_bool (client, proxy_key, NULL)) {
		proxyhost = gconf_client_get_string (client, phost_key, NULL);
		proxyport = gconf_client_get_int (client, pport_key, NULL);
		proxyurl = g_strdup_printf ("http://%s:%d", proxyhost,
					    proxyport);
		gnome_transfer_proxy (tran, proxyurl);
		if (gconf_client_get_bool (client, pauth_key, NULL)) {
			proxyuser = gconf_client_get_string (client, puser_key,
							     NULL);
			proxypass = gconf_client_get_string (client, ppass_key,
							     NULL);
			gnome_transfer_proxy_password (tran, proxyuser,
						       proxypass);
		}
	}

	if (gnome_transfer_begin (tran, url) == -1) {
		char *txt;
		GtkWidget *dlg;

		txt = g_strdup_printf (_("Connection failed: %s"), 
				       gnome_transfer_reason (tran));
		dlg = gnome_error_dialog (txt);
		g_free (txt);
		gnome_dialog_run_and_close (GNOME_DIALOG (dlg));
		exit (1);
	}

	g_free (proxyhost);
	g_free (proxyurl);
	g_free (proxyuser);
	g_free (proxypass);
	g_free (url);
}

static void
chanlist_transfer_done_cb (GnomeTransfer * tran,
			   GnomeTransferStatus status, 
			   struct server *serv)
{
	gchar **lines = NULL, **iter;
	ChannelInfo *info = NULL;

	if (status == GNOME_TRANSFER_FAILED) goto ERROR;
	if (serv->gui->channel_response_buf->len == 0) goto ERROR;

	g_byte_array_append (serv->gui->channel_response_buf, "\0", 1);

	lines = g_strsplit (serv->gui->channel_response_buf->data, "\n", 0);

	if (!lines) goto ERROR;
	else if (!lines [0] || !lines [1]) {
		g_strfreev (lines);
		goto ERROR;
	}

	info = g_new0 (ChannelInfo, 1);
	info->serv = serv;

	for (iter = lines; *iter; iter++) {
		gchar *colon, *key, *val;

		if (iter [0][0] == '\0') {
			if (info &&
			    info->name &&
			    info->desc &&
			    info->channel)
				serv->gui->channel_info = 
					g_slist_append (
						serv->gui->channel_info, 
						info);
			else g_free (info);

			if (!iter [1]) break;

			info = g_new0 (ChannelInfo, 1);
			info->serv = serv;
			continue;
		}

		colon = strchr (*iter, ':');
		if (!colon) break;

		*colon = '\0';

		key = g_strstrip (*iter);
		val = g_strstrip (colon + 1);

		if (!g_strcasecmp (key, "name"))
			info->name = g_strdup (val);
		else if (!g_strcasecmp (key, "description"))
			info->desc = g_strdup (val);
		else if (!g_strcasecmp (key, "channel"))
			info->channel = g_strdup (val);
		else if (!g_strcasecmp (key, "server")) {
			colon = strchr (val, ':');
			if (colon) {
				info->port = atoi (colon + 1);
				*colon = '\0';
			}
			info->server = g_strdup (val);
		}
	}

	g_byte_array_free (serv->gui->channel_response_buf, TRUE);
	g_strfreev (lines);

	return;

 ERROR:
	gnome_dialog_run_and_close (
		gtkutil_simpledialog ("Unable to contact Ximian Support "
				      "server.\n"));

	exit (1);	
}

static void
begin_channel_list_query (struct server *serv)
{
	gchar *url, *proxyurl = NULL;
	gchar *proxyhost = NULL;
	gint proxyport = 8080;
	gchar *proxyuser = NULL, *proxypass = NULL;
	GConfClient *client;

	GnomeTransfer *tran = gnome_transfer_new ();

	gtk_signal_connect (GTK_OBJECT (tran),
			    "file_data",
			    GTK_SIGNAL_FUNC (transfer_data_cb), 
			    &serv->gui->channel_response_buf);

	gtk_signal_connect (GTK_OBJECT (tran),
			    "file_done",
			    GTK_SIGNAL_FUNC (chanlist_transfer_done_cb), 
			    serv);

	url = construct_url ("http://monkeytalk.ximian.com/cgi-bin/list-channels");

	client = gconf_client_get_default ();

	if (gconf_client_get_bool (client, proxy_key, NULL)) {
		proxyhost = gconf_client_get_string (client, phost_key, NULL);
		proxyport = gconf_client_get_int (client, pport_key, NULL);
		proxyurl = g_strdup_printf ("http://%s:%d", proxyhost,
					    proxyport);
		gnome_transfer_proxy (tran, proxyurl);
		if (gconf_client_get_bool (client, pauth_key, NULL)) {
			proxyuser = gconf_client_get_string (client, puser_key,
							     NULL);
			proxypass = gconf_client_get_string (client, ppass_key,
							     NULL);
			gnome_transfer_proxy_password (tran, proxyuser,
						       proxypass);
		}
	}

	if (gnome_transfer_begin (tran, url) == -1) {
		g_warning ("Download of channel information failed: %s",
			   gnome_transfer_reason (tran));
	}

	g_free (proxyhost);
	g_free (proxyurl);
	g_free (proxyuser);
	g_free (proxypass);
	g_free (url);
}

static void
connect_to_support_channel (struct session *sess)
{
	static gboolean has_connected = 0;
	if (has_connected) 
		return;
	has_connected = 1;
	
	begin_default_channel_query (sess);
	
	while (gtk_events_pending ()) gtk_main_iteration ();
	
	if (support_greeting) {
		char **text, **iter;
		
		text = g_strsplit (support_greeting, "\n\n", 0);
		
		for (iter = text; *iter; iter++) {
			gtk_xtext_append_indent (
				GTK_XTEXT (sess->gui->textgad), 
				NULL,
				0,
				*iter, 
				strlen (*iter));
			
			gtk_xtext_append_indent (
				GTK_XTEXT (sess->gui->textgad), 
				NULL,
				0,
				NULL, 
				0);
		}
		
		g_strfreev (text);
	}
	
	begin_channel_list_query (sess->server);
}

void 
fe_new_window (struct session *sess)
{
	create_talk_window_new (sess);
}

void 
fe_new_server (struct server *serv)
{
	serv->gui = g_new0 (TalkServer, 1);
}

void 
fe_message (char *msg, int wait)
{
	GtkWidget *dialog;

	dialog = gtkutil_simpledialog (msg);
	if (wait) {
		gtk_signal_connect (GTK_OBJECT (dialog), "destroy",
				    gtkutil_null_this_var, &dialog);

		while (dialog) gtk_main_iteration ();
	}
}

void 
fe_input_remove (int tag)
{
	gdk_input_remove (tag);
}

int 
fe_input_add (int sok, int read, int write, int ex, void *func, void *data)
{
	int type = 0;

	if (read) type |= GDK_INPUT_READ;
	if (write) type |= GDK_INPUT_WRITE;
	if (ex) type |= GDK_INPUT_EXCEPTION;

	return gdk_input_add (sok, type, func, data);
}

void 
fe_set_topic (struct session *sess, char *topic)
{
	char tbuf[200];

	if (sess->is_dialog)
		snprintf (tbuf, 
			  sizeof (tbuf), 
			  "MonkeyTalk: Private Chat with %s", 
			  topic);
	else {
		snprintf (tbuf, sizeof (tbuf), "MonkeyTalk: %s", topic);

		gnome_canvas_item_set (sess->gui->banner,
				       "text", topic,
				       NULL);
	}

	gtk_window_set_title (GTK_WINDOW (sess->gui->window), tbuf);
}

void 
fe_set_hilight (struct session *sess)
{
}

void 
fe_update_channel_key (struct session *sess)
{
}

void 
fe_update_channel_limit (struct session *sess)
{
}

int 
fe_is_chanwindow (struct server *serv)
{
	if (!serv->gui->chanlist_window) return 0;
	return 1;
}

void 
fe_chan_list_end (struct server *serv)
{
}

void 
fe_notify_update (char *name)
{
}

void 
fe_text_clear (struct session *sess)
{
	GtkXText *xtext = GTK_XTEXT (sess->gui->textgad);
	gtk_xtext_clear (xtext);
	if (xtext->auto_indent) xtext->indent = 1;
}

void 
fe_close_window (struct session *sess)
{
	gtk_widget_destroy (sess->gui->window);
}

static int 
update_progressbar (struct session *sess)
{
	static int type = 0;
	static float pos = 0;

	if (!is_session (sess))
		return 0;

	pos += 0.05;
	if (pos >= 0.99) {
		if (type == 0) {
			type = 1;
			gtk_progress_bar_set_orientation ((GtkProgressBar *)
							  sess->gui->bar,
							  GTK_PROGRESS_RIGHT_TO_LEFT);
		} else {
			type = 0;
			gtk_progress_bar_set_orientation ((GtkProgressBar *)
							  sess->gui->bar,
							  GTK_PROGRESS_LEFT_TO_RIGHT);
		}
		pos = 0.05;
	}
	gtk_progress_bar_update ((GtkProgressBar *) sess->gui->bar, pos);
	return 1;
}

void 
fe_progressbar_start (struct session *sess)
{
	if (!sess->gui->op_box) return;

	sess->gui->bar = gtk_progress_bar_new ();
	gtk_box_pack_start (GTK_BOX (sess->gui->op_box), 
			    sess->gui->bar,
			    0, 
			    0, 
			    0);
	gtk_widget_show (sess->gui->bar);
	sess->server->bartag =
		gtk_timeout_add (50, (GtkFunction) update_progressbar, sess);
}

void 
fe_progressbar_end (struct session *sess)
{
	struct server *serv;
	GSList *list = sess_list;

	if (!sess) return;

	serv = sess->server;
	while (list) {	
		/* check all windows that use this server and  *
		 * remove the connecting graph, if it has one. */
		sess = (struct session *) list->data;
		if (sess->server == serv && sess->gui->bar) {
			if (GTK_IS_WIDGET (sess->gui->bar))
				gtk_widget_destroy (sess->gui->bar);
			sess->gui->bar = 0;
			gtk_timeout_remove (sess->server->bartag);
		}
		list = list->next;
	}
}

void 
fe_print_text (struct session *sess, char *text)
{
	PrintTextRaw (sess->gui->textgad, text, TRUE);

	/* highlight window here? */
}

void 
fe_beep (void)
{
	gdk_beep ();
}

int 
fe_is_confmode (struct session *sess)
{
	return 0;
}

void 
fe_lastlog (session * sess, session * lastlog_sess, char *sstr)
{
	textentry *ent;

	ent = GTK_XTEXT (sess->gui->textgad)->text_first;
	if (!ent) {
		PrintText (lastlog_sess, "Search buffer is empty.\n");
	} else {
		do {
			if (nocasestrstr (ent->str, sstr))
				PrintText (lastlog_sess, ent->str);
			ent = ent->next;
		}
		while (ent);
	}
}

void 
fe_set_lag (server * serv, int lag)
{
}

/* nick change for a user we are messageing? */
void 
fe_change_nick (struct server *serv, char *nick, char *newnick)
{
	struct session *sess = find_dialog (serv, nick);
	if (sess) {
		strcpy (sess->channel, newnick);
		gtk_window_set_title (GTK_WINDOW (sess->gui->window), newnick);
	}
}

void 
fe_ignore_update (int level)
{
}

/* update user icon (next to nick)? */
void 
fe_set_away (server * serv)
{
}

void 
fe_update_mode_buttons (struct session *sess, char mode, char sign)
{
}

void 
fe_set_channel (struct session *sess)
{
}

void 
fe_set_nonchannel (struct session *sess, int state)
{
}

void 
fe_clear_channel (struct session *sess)
{
	clear_user_list (sess);

	/* 
	 * Check me. set_title might be called after clear_channel by common.
	 * maybe we should just close the window here?
	 */
	gnome_canvas_item_set (sess->gui->banner,
                               "text", "",
                               NULL);
	gtk_window_set_title (GTK_WINDOW (sess->gui->window), "MonkeyTalk");

	if (sess->gui->op_xpm) gtk_widget_destroy (sess->gui->op_xpm);
	sess->gui->op_xpm = NULL;
}

void 
fe_set_nick (struct server *serv, char *newnick)
{
	GSList *list = sess_list;
	struct session *sess;

	strcpy (serv->nick, newnick);
	while (list) {
		sess = (struct session *) list->data;
		if (sess->server == serv)
			gtk_label_set_text (GTK_LABEL (sess->gui->nickname),
					    newnick);
		list = list->next;
	}
}

void 
fe_buttons_update (struct session *sess)
{
}

void 
fe_session_callback (struct session *sess)
{
	if (sess->gui->bar) {
		fe_progressbar_end (sess);
		sess->server->connecting = TRUE;
	}

	if (menu_sess == sess && sess_list)
		menu_sess = (struct session *) sess_list->data;

	gnome_config_set_bool ("/monkeytalk/Default/hideuserlist",
			       prefs.hideuserlist);
	gnome_config_sync ();
}

void 
fe_add_rawlog (struct server *serv, char *text, int outbound)
{
}

void 
fe_checkurl (char *buf)
{
}
